# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

# https://en.wikipedia.org/wiki/Comparison_of_executable_file_formats
# https://github.com/erocarrera/pefile/blob/master/pefile.py#L2914
# pefile only handles MZ magic bytes, but ZM might be valid as well
# there could also be some other supported Windows EXE formats such as NE, LE, LX, TX, and COM (generally no header, except CP/M 3 format COM has RET instruction)

import pathlib
import re
from typing import Any, Dict, List, Optional

import defusedxml.ElementTree
import dnfile
from loguru import logger

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


def supports_file(filetype) -> bool:
    return filetype == "PE"


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: str) -> object:
    if not supports_file(filetype):
        return None
    return extract_pe_info(filename)


# https://learn.microsoft.com/en-us/windows/win32/debug/pe-format#machine-types
# Values for CPU types that can appear in a PE file
pe_machine_types = {
    0x0: "UNKNOWN",
    0x184: "ALPHA",  # Alpha_AXP
    0x284: "ALPHA64",  # Also AXP64
    0x1D3: "AM33",
    0x8664: "AMD64",
    0x1C0: "ARM",
    0xAA64: "ARM64",
    0x1C4: "ARMNT",
    0xC0EE: "CEE",
    0xCEF: "CEF",
    0xEBC: "EBC",
    0x14C: "I386",
    0x200: "IA64",
    0x6232: "LOONGARCH32",
    0x6264: "LOONGARCH64",
    0x9041: "M32R",
    0x266: "MIPS16",
    0x366: "MIPSFPU",
    0x466: "MIPSFPU16",
    0x1F0: "POWERPC",
    0x1F1: "POWERPCFP",
    0x160: "R3000",  # Big-Endian
    0x162: "R3000",  # Little-Endian
    0x166: "R4000",
    0x168: "R10000",
    0x5032: "RISCV32",
    0x5064: "RISCV64",
    0x5128: "RISCV128",
    0x1A2: "SH3",
    0x1A3: "SH3DSP",
    0x1A4: "SH3E",
    0x1A6: "SH4",
    0x1A8: "SH5",
    0x1C2: "THUMB",
    0x520: "TRICORE",
    0x169: "WCEMIPSV2",
}

# https://learn.microsoft.com/en-us/windows/win32/debug/pe-format#windows-subsystem
# Values for Windows subsystem optional header field that determines which is required to run the image
pe_subsystem_types = {
    0: "UNKNOWN",
    1: "NATIVE",
    2: "WINDOWS_GUI",
    3: "WINDOWS_CUI",
    5: "OS2_CUI",
    7: "POSIX_CUI",
    8: "NATIVE_WINDOWS",
    9: "WINDOWS_CE_GUI",
    10: "EFI_APPLICATION",
    11: "EFI_BOOT_SERVICE_DRIVER",
    12: "EFI_RUNTIME_DRIVER",
    13: "EFI_ROM",
    14: "XBOX",
    16: "WINDOWS_BOOT_APPLICATION",
    17: "XBOX_CODE_CATALOG",
}


def extract_pe_info(filename: str) -> object:
    try:
        pe = dnfile.dnPE(filename, fast_load=False)
    except (OSError, dnfile.PEFormatError):
        return {}

    file_details: Dict[str, Any] = {"OS": "Windows"}
    if pe.FILE_HEADER is not None:
        if pe.FILE_HEADER.Machine in pe_machine_types:
            file_details["peMachine"] = pe_machine_types[pe.FILE_HEADER.Machine]
        else:
            file_details["peMachine"] = pe.FILE_HEADER.Machine
            logger.warning("Unknown machine type encountered in PE file header")
    if pe.OPTIONAL_HEADER is not None:
        file_details["peOperatingSystemVersion"] = (
            f"{pe.OPTIONAL_HEADER.MajorOperatingSystemVersion}.{pe.OPTIONAL_HEADER.MinorOperatingSystemVersion}"
        )
        file_details["peSubsystemVersion"] = (
            f"{pe.OPTIONAL_HEADER.MajorSubsystemVersion}.{pe.OPTIONAL_HEADER.MinorSubsystemVersion}"
        )
        if pe.OPTIONAL_HEADER.Subsystem in pe_subsystem_types:
            file_details["peSubsystem"] = pe_subsystem_types[pe.OPTIONAL_HEADER.Subsystem]
        else:
            file_details["peSubsystem"] = pe.OPTIONAL_HEADER.Subsystem
            logger.warning("Unknown Windows Subsystem type encountered in PE file header")
        file_details["peLinkerVersion"] = (
            f"{pe.OPTIONAL_HEADER.MajorLinkerVersion}.{pe.OPTIONAL_HEADER.MinorLinkerVersion}"
        )

    if import_dir := getattr(pe, "DIRECTORY_ENTRY_IMPORT", None):
        # Imported Symbols
        file_details["peImport"] = []
        for entry in import_dir:
            file_details["peImport"].append(entry.dll.decode())

    if bound_import_dir := getattr(pe, "DIRECTORY_ENTRY_BOUND_IMPORT", None):
        # Bound Imported Symbols
        file_details["peBoundImport"] = []
        for entry in bound_import_dir:
            file_details["peBoundImport"].append(entry.name.decode())

    if delay_import_dir := getattr(pe, "DIRECTORY_ENTRY_DELAY_IMPORT", None):
        # Delay Imported Symbols
        file_details["peDelayImport"] = []
        for entry in delay_import_dir:
            file_details["peDelayImport"].append(entry.dll.decode())

    file_details["peIsExe"] = pe.is_exe()
    file_details["peIsDll"] = pe.is_dll()
    if opt_hdr := getattr(pe, "OPTIONAL_HEADER", None):
        if opt_hdr_data_dir := getattr(opt_hdr, "DATA_DIRECTORY", None):
            # COM Descriptor, used to identify CLR/.NET binaries
            com_desc_dir_num = dnfile.DIRECTORY_ENTRY["IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR"]
            com_desc_dir = opt_hdr_data_dir[com_desc_dir_num]
            file_details["peIsClr"] = (com_desc_dir.VirtualAddress > 0) and (com_desc_dir.Size > 0)

    if pe_fi := getattr(pe, "FileInfo", None):
        if len(pe_fi) > 0:
            file_details["FileInfo"] = {}
            for fi_entry in pe_fi[0]:
                if fi_entry.name == "StringFileInfo" and hasattr(fi_entry, "StringTable"):
                    for st in fi_entry.StringTable:
                        for st_entry in st.entries.items():
                            file_details["FileInfo"][st_entry[0].decode()] = st_entry[1].decode()

    # If this is a .NET assembly, extract information about it from the headers
    if dnet := getattr(pe, "net", None):
        if dnet_flags := getattr(dnet, "Flags", None):
            file_details["dotnetFlags"] = {
                "ILONLY": dnet_flags.CLR_ILONLY,
                "32BITREQUIRED": dnet_flags.CLR_32BITREQUIRED,
                "IL_LIBRARY": dnet_flags.CLR_IL_LIBRARY,
                "STRONGNAMESIGNED": dnet_flags.CLR_STRONGNAMESIGNED,
                "NATIVE_ENTRYPOINT": dnet_flags.CLR_NATIVE_ENTRYPOINT,
                "TRACKDEBUGDATA": dnet_flags.CLR_TRACKDEBUGDATA,
                "32BITPREFERRED": dnet_flags.CLR_PREFER_32BIT,
            }
        if dnet_mdtables := getattr(dnet, "mdtables", None):
            if assembly_info := getattr(dnet_mdtables, "Assembly", None):
                assemblies = []
                for a_info in assembly_info:
                    assemblies.append(get_assembly_info(a_info))
                file_details["dotnetAssembly"] = assemblies
            if assemblyref_info := getattr(dnet_mdtables, "AssemblyRef", None):
                assembly_refs = []
                for ar_info in assemblyref_info:
                    assembly_refs.append(get_assemblyref_info(ar_info))
                file_details["dotnetAssemblyRef"] = assembly_refs
            if implmap_info := getattr(dnet_mdtables, "ImplMap", None):
                imp_modules: List[Dict[str, Any]] = []
                for im_info in implmap_info:
                    insert_implmap_info(im_info, imp_modules)
                file_details["dotnetImplMap"] = imp_modules

    # TODO for a custom intermediate SBOM format, the information read from the manifest and app config files
    # should be tied to a specific "<install path>/<file name>", in case the same file appears in separate
    # directories/file systems with different manifest/config file settings for paths to search
    file_details["dllRedirectionLocal"] = check_windows_dll_redirection_local(filename)

    manifest_info = get_windows_manifest_info(filename)
    if manifest_info:
        file_details["manifestFile"] = manifest_info

    app_config_info = get_windows_application_config_info(filename)
    if app_config_info:
        file_details["appConfigFile"] = app_config_info

    return file_details


def add_core_assembly_info(asm_dict: Dict[str, Any], asm_info):
    # REFERENCE: https://github.com/malwarefrank/dnfile/blob/096de1b3/src/dnfile/stream.py#L36-L39
    # HeapItemString value will be decoded string, or None if there was a UnicodeDecodeError
    asm_dict["Name"] = asm_info.Name.value if asm_info.Name.value else asm_info.raw_data.hex()
    asm_dict["Culture"] = (
        asm_info.Culture.value if asm_info.Culture.value else asm_info.Culture.raw_data.hex()
    )
    asm_dict["Version"] = (
        f"{asm_info.MajorVersion}.{asm_info.MinorVersion}.{asm_info.BuildNumber}.{asm_info.RevisionNumber}"
    )
    # REFERENCE: https://github.com/malwarefrank/dnfile/blob/096de1b3/src/dnfile/stream.py#L62-L66
    # HeapItemBinary value is the bytes following the compressed int (indicating the length)
    if asm_info.PublicKey is not None:
        asm_dict["PublicKey"] = (
            # raw_data attribute of PublicKey includes leading byte with length of data, value attr removes it
            asm_info.PublicKey.value.hex()
        )


def add_assembly_flags_info(asm_dict, asm_info):
    # Info on flags
    # Processor Architecture fields/values: https://learn.microsoft.com/en-us/dotnet/api/system.reflection.processorarchitecture?view=net-6.0
    # PA enum values in dnfile: https://github.com/malwarefrank/dnfile/blob/7441fe326e0cc254ed2944a18773d8b4fe99c4c6/src/dnfile/enums.py#L663-L671
    # .NET runtime corhdr.h PA enum: https://github.com/dotnet/runtime/blob/9d6396deb02161f5ee47af72ccac52c2e1bae458/src/coreclr/inc/corhdr.h#L747-L754
    # Assembly flags: https://learn.microsoft.com/en-us/dotnet/api/system.reflection.assemblyflags?view=net-6.0
    if a_flags := getattr(asm_info, "Flags", None):
        asm_dict["Flags"] = {
            "DisableJitCompileOptimizer": a_flags.afDisableJITcompileOptimizer,  # JIT compiler optimization disabled for assembly
            "EnableJitCompileTracking": a_flags.afEnableJITcompileTracking,  # JIT compiler tracking enabled for assembly
            "PublicKey": a_flags.afPublicKey,  # assembly ref has the full (unhashed) public key
            "Retargetable": a_flags.afRetargetable,  # impl of referenced assembly used at runtime may not match version seen at compile time
            "PA_Specified": a_flags.afPA_Specified,  # propagate processor architecture flags to AssemblyRef record
            "PA_None": a_flags.afPA_None,
            "PA_MSIL": a_flags.afPA_MSIL,
            "PA_x86": a_flags.afPA_x86,
            "PA_IA64": a_flags.afPA_IA64,
            "PA_AMD64": a_flags.afPA_AMD64,
            "PA_ARM": a_flags.afPA_Unknown1,  # based on enumeration values in docs, clr runtime corhdr.h, and dnfile
            "PA_ARM64": a_flags.afPA_Unknown2,
            "PA_NoPlatform": a_flags.afPA_Unknown3,  # applies to any platform but cannot run on any (e.g. reference assembly), "specified" should not be set
        }


def get_assembly_info(asm_info) -> Dict[str, Any]:
    asm: Dict[str, Any] = {}
    add_core_assembly_info(asm, asm_info)
    # REFERENCE: https://github.com/malwarefrank/dnfile/blob/fcccdaf/src/dnfile/enums.py#L851-L863
    # HashAlgID is a dnfile enum, based on possible .NET hash algs
    asm["HashAlgId"] = asm_info.HashAlgId.name
    add_assembly_flags_info(asm, asm_info)
    return asm


def get_assemblyref_info(asmref_info) -> Dict[str, Any]:
    asmref: Dict[str, Any] = {}
    add_core_assembly_info(asmref, asmref_info)
    # REFERENCE: https://github.com/malwarefrank/dnfile/blob/096de1b3/src/dnfile/stream.py#L62-L66
    # HeapItemBinary value is the bytes following the compressed int (indicating the length)
    # raw_data attribute has the compressed int indicating length included
    asmref["HashValue"] = asmref_info.HashValue.value.hex()
    add_assembly_flags_info(asmref, asmref_info)
    return asmref


def insert_implmap_info(im_info, imp_modules: List[Dict[str, Any]]):
    # REFERENCE: https://github.com/malwarefrank/dnfile/blob/096de1b3/src/dnfile/stream.py#L36-L39
    # HeapItemString value will be decoded string, or None if there was a UnicodeDecodeError
    dllName = (
        im_info.ImportScope.row.Name.value
        if im_info.ImportScope.row.Name.value
        else im_info.ImportScope.row.Name.raw_data.hex()
    )
    methodName = (
        im_info.ImportName.value if im_info.ImportName.value else im_info.ImportName.raw_data.hex()
    )
    if dllName:
        for imp_module in imp_modules:
            if imp_module["Name"] == dllName:
                imp_module["Functions"].append(methodName)
                return
        imp_modules.append({"Name": dllName, "Functions": [methodName]})


def get_xmlns_and_tag(uri):
    check_xmlns = re.match(r"\{.*\}", uri.tag)
    xmlns = check_xmlns.group(0) if check_xmlns else ""
    tag = uri.tag
    if tag.startswith(xmlns):
        tag = tag.replace(xmlns, "", 1)
    return xmlns, tag


# check for manifest file on Windows (note: could also be a resource contained within an exe/dll)
# return any info that could be useful for establishing "Uses" relationships later
def get_windows_manifest_info(filename: str) -> Optional[Dict[str, Any]]:
    binary_filepath = pathlib.Path(filename)
    manifest_filepath = binary_filepath.with_suffix(binary_filepath.suffix + ".manifest")
    if manifest_filepath.exists():
        logger.info("Found application manifest file for " + filename)
        et = defusedxml.ElementTree.parse(manifest_filepath)
        manifest_info = {}

        # root element is <assembly> which could contain:
        # <assemblyIdentity>
        # <file>
        # <dependency>, which contains the usual <dependentAssembly> element found in Windows app config files
        for asm_e in et.getroot():
            asm_xmlns, asm_tag = get_xmlns_and_tag(asm_e)
            if asm_tag == "assemblyIdentity":
                if "assemblyIdentity" in manifest_info:
                    logger.warning(
                        "duplicate assemblyIdentity element found in the manifest file: "
                        + str(manifest_filepath)
                    )
                manifest_info["assemblyIdentity"] = asm_e.attrib
            if asm_tag == "file":
                if "file" not in manifest_info:
                    manifest_info["file"] = []
                manifest_info["file"].append(asm_e.attrib)
            if asm_tag == "dependency":
                if "dependency" in manifest_info:
                    logger.warning(
                        "duplicate dependency element found in the manifest file: "
                        + str(manifest_filepath)
                    )
                dependency_info: Dict[str, Any] = {}
                for dependency in asm_e:
                    dependency_xmlns, dependency_tag = get_xmlns_and_tag(dependency)
                    if dependency_tag == "dependentAssembly":
                        if "dependentAssembly" not in dependency_info:
                            dependency_info["dependentAssembly"] = []
                        dependency_info["dependentAssembly"].append(
                            get_dependentAssembly_info(dependency)
                        )
                manifest_info["dependency"] = dependency_info
        return manifest_info
    return None


# returns info on a dependentAssembly
def get_dependentAssembly_info(da_et, config_filepath=""):
    daet_xmlns, daet_tag = get_xmlns_and_tag(da_et)
    if daet_tag != "dependentAssembly":
        logger.warning("element tree given was not for a dependentAssembly element tag")
    da_info = {}
    for da_e in da_et:
        da_xmlns, da_tag = get_xmlns_and_tag(da_e)
        if da_tag == "assemblyIdentity":
            if "assemblyIdentity" in da_info:
                logger.warning(
                    "duplicate assemblyIdentity element found in the app config file: "
                    + str(config_filepath)
                )
            da_info["assemblyIdentity"] = da_e.attrib
        if da_tag == "codeBase":
            if "codeBase" in da_info:
                logger.warning(
                    "duplicate codeBase element found in the app config file: "
                    + str(config_filepath)
                )
            da_info["codeBase"] = da_e.attrib
        if da_tag == "bindingRedirect":
            if "bindingRedirect" in da_info:
                logger.warning(
                    "duplicate bindingRedirect element found in the app config file: "
                    + str(config_filepath)
                )
            da_info["bindingRedirect"] = da_e.attrib
    return da_info


# returns a map for the given assembly binding element tree based on content within the element tree for an "assemblyBinding" tag
# the "assemblyBinding" tag can appear within either a <runtime> or <windows> element, or under the root <configuration> element
# <runtime>: could contain appliesTo attribute, probing, dependentAssembly, and qualifyAssembly elements
# <windows>: could contain probing, assemblyIdentity, and dependentAssembly elements
def get_assemblyBinding_info(ab_et, config_filepath=""):
    xmlns, tag = get_xmlns_and_tag(ab_et)
    if tag != "assemblyBinding":
        logger.warning("element tree given was not for an assemblyBinding tag")

    ab_info = {}

    # Specifies runtime version .NET assembly redirection applies to
    # uses .NET Framework version number; if not given, assemblyBinding
    # element applies to all versions of .NET Framework
    # ab_e.attrib["appliesTo"]
    if "appliesTo" in ab_et.attrib:
        ab_info["appliesTo"] = ab_et.attrib["appliesTo"]
    for ab_e in ab_et:
        ab_xmlns, ab_tag = get_xmlns_and_tag(ab_e)
        # <probing>
        # specifies subdirs of application's base dir to search for assemblies
        # privatePath: "bin;bin2\subbin;bin3"
        if ab_tag == "probing":
            if "probing" in ab_info:
                logger.warning(
                    "duplicate probing element found in the app config file: "
                    + str(config_filepath)
                )
            ab_info["probing"] = ab_e.attrib

        # <dependentAssembly> for .NET
        # binding policy and assembly location for dependent assemblies
        #   <assemblyIdentity>
        #   used to determine if this dependentAssembly config element should apply
        #   - name: name of the assembly
        #   - culture: (optional) specify language and country/region of assembly
        #   - publicKeyToken: (optional) specify assembly strong name
        #   - processorArchitecture: (optional) "x86", "amd64", "msil", or "ia64"
        #   <codeBase>
        #   specifies assembly to use; if not present, usual probing for assemblies
        #   - version: version of assembly the codebase applies to
        #   - href: URL where runtime can find specified version of assembly
        #   <bindingRedirect>
        #   redirect one assembly to another version
        #   - oldVersion: assembly version originally requested
        #   - newVersion: the assembly version to use instead
        if ab_tag == "dependentAssembly":
            if "dependentAssembly" not in ab_info:
                ab_info["dependentAssembly"] = []
            ab_info["dependentAssembly"].append(get_dependentAssembly_info(ab_e, config_filepath))

        # <qualifyAssembly>
        # replaces partial name in Assembly.Load with full name
        # - partialName: "math"
        # - fullName: "math,version=...,publicKeyToken=...,culture=neutral"
        if ab_tag == "qualifyAssembly":
            if "qualifyAssembly" in ab_info:
                logger.warning(
                    "duplicate qualifyAssembly element found in the app config file: "
                    + str(config_filepath)
                )
            ab_info["qualifyAssembly"] = ab_e.attrib
    return ab_info


# DLL redirection summary: redirection file with name_of_exe.local (contents are ignored) makes a check for mydll.dll happen in the application directory first,
# regardless of what the full path specified for LoadLibrary or LoadLibraryEx is (if no dll found in local directory, uses the typical search order)
def check_windows_dll_redirection_local(filename: str):
    binary_filepath = pathlib.Path(filename)
    config_filepath = binary_filepath.with_suffix(binary_filepath.suffix + ".local")
    return config_filepath.exists()


# check for an application configuration file and return (potentially) useful information
# https://learn.microsoft.com/en-us/dotnet/framework/deployment/how-the-runtime-locates-assemblies#application-configuration-file
# https://learn.microsoft.com/en-us/windows/win32/sbscs/application-configuration-files
def get_windows_application_config_info(filename: str):
    binary_filepath = pathlib.Path(filename)
    config_filepath = binary_filepath.with_suffix(binary_filepath.suffix + ".config")
    if config_filepath.exists():
        logger.info("Found application configuration file for " + filename)
        et = defusedxml.ElementTree.parse(config_filepath)
        app_config_info = {}

        # requiredRuntime is used for v1.0 of .NET Framework, supportedRuntime is for v1.1+
        supportedRuntime = et.find("./startup/supportedRuntime")
        requiredRuntime = et.find("./startup/requiredRuntime")
        if (supportedRuntime is not None) or (requiredRuntime is not None):
            startup_info = {}
            if (supportedRuntime is not None) and supportedRuntime.attrib:
                startup_info["supportedRuntime"] = supportedRuntime.attrib
            if (requiredRuntime is not None) and requiredRuntime.attrib:
                startup_info["requiredRuntime"] = requiredRuntime.attrib
            app_config_info["startup"] = startup_info

        # <linkedConfiguration href="URL of linked XML file" />
        # - seems to appear within an assemblyBinding element right under the root configuration element
        # - only format for href is `file://` either local or UNC
        # - includes assembly config file contents here, similar to #include
        linkedConfiguration = et.find("./assemblyBinding/linkedConfiguration")
        if (linkedConfiguration is not None) and linkedConfiguration.attrib:
            app_config_info["assemblyBinding"] = {"linkedConfiguration": linkedConfiguration.attrib}

        # The following appear within a <windows> element:
        # <probing privatePath="bin;..\bin2\subbin;bin3"/>
        # - valid starting on Windows Server 2008 R2 and Windows 7
        # - privatePath: (optional) specifies relative paths of subdirs of the app base dir that might contain assemblies
        # <assemblyBinding>
        # - first element must be an assemblyIdentity describing the assembly; followed by dependentAssembly elements (also below)
        #   <probing privatePath="bin;..\bin2\subbin;bin3"/>
        #   - valid starting on Windows Server 2008 R2 and Windows 7
        #   - privatePath: (optional) specifies relative paths of subdirs of the app base dir that might contain assemblies
        # <dependency> (based on the app config file schema, this
        # - contains one or more dependentAssembly elements (optional)
        #  <dependentAssembly>
        #  - contains assemblyIdentity that unique identifies an app
        #  - app config file redirects binding of application to side-by-side assemblies
        #    <assemblyIdentity processorArchitecture="X86" name="Microsoft.Windows.mysampleApp" type="win32" version="1.0.0.0"/>
        #    - describes side-by-side assembly the application depends on
        #    - type: must be "win32" lowercase
        #    - name: identifies app being affected by app config file or assembly being redirected
        #    - language: (optional) identifies the language by DHTML code or "*" if language neutral/worldwide use
        #    - processorArchitecture: the processor running the application
        #    - version: version of app or assembly
        #    - publicKeyToken: 16-char hex string w/ last 8 bytes of SHA-1 hash of public key the assembly is signed by
        #    <bindingRedirect oldVersion="1.0.0.0" newVersion="1.0.10.0"/>
        #    - oldVersion: assembly version being overridden or redirected
        #    - newVersion: replacement assembly version
        windows_et = et.find("./windows")
        if windows_et is not None:
            windows_info = {}
            for win_child in windows_et:
                xmlns, tag = get_xmlns_and_tag(win_child)
                if tag == "probing":
                    if "probing" in windows_info:
                        logger.warning(
                            "duplicate windows/probing element was found in the app config file: "
                            + str(config_filepath)
                        )
                    if "privatePath" in win_child.attrib:
                        windows_info["probing"] = {"privatePath": win_child.attrib["privatePath"]}
                    else:
                        logger.warning(
                            "windows/probing element missing privatePath attribute in app config file: "
                            + str(config_filepath)
                        )
                if tag == "assemblyBinding":
                    windows_info["assemblyBinding"] = get_assemblyBinding_info(win_child)
                if tag == "dependency":
                    dependency_info: Dict[str, Any] = {}
                    for dependency in win_child:
                        dependency_xmlns, dependency_tag = get_xmlns_and_tag(dependency)
                        if dependency_tag == "dependentAssembly":
                            if "dependentAssembly" not in dependency_info:
                                dependency_info["dependentAssembly"] = []
                            dependency_info["dependentAssembly"].append(
                                get_dependentAssembly_info(dependency, config_filepath)
                            )
                    windows_info["dependency"] = dependency_info
            app_config_info["windows"] = windows_info

        # runtime element used for .NET related configuration info that can affect how the runtime locates assemblies to load
        runtime_et = et.find("./runtime")
        if runtime_et is not None:
            runtime_info = {}
            for rt_child in runtime_et:
                # Docs say "urn:schemas-microsoft-com:asm.v1" is the namespace for assemblyBinding
                xmlns, tag = get_xmlns_and_tag(rt_child)
                if tag == "developmentMode":
                    # attribute is either 'true' or 'false' (string)
                    # Causes runtime to search directory given in DEVPATH env var for assemblies first (skips signature checks)
                    if "developmentMode" in runtime_info:
                        logger.warning(
                            "duplicate developmentMode element was found in the app config file: "
                            + str(config_filepath)
                        )
                    if "developerInstallation" in rt_child.attrib:
                        runtime_info["developmentMode"] = {
                            "developerInstallation": rt_child.attrib["developerInstallation"]
                        }
                    else:
                        logger.warning(
                            "developmentMode element missing developerInstallation attribute in app config file: "
                            + str(config_filepath)
                        )
                if tag == "assemblyBinding":
                    runtime_info["assemblyBinding"] = get_assemblyBinding_info(
                        rt_child, config_filepath
                    )
            app_config_info["runtime"] = runtime_info

        # Info returned includes:
        # - runtime information related to .NET
        # - config options (codeBase, probing) that affect how .NET assemblies are located
        # - config options (probing) that affect how native Windows DLLs are found
        # - assembly identity info (processor targeted, etc)
        # - dependent assembly info (used to determine correct version of an assembly to use)
        return app_config_info
    return None



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from typing import Optional

import surfactant.plugin
from surfactant.sbomtypes import SBOM


@surfactant.plugin.hookimpl
def read_sbom(infile) -> SBOM:
    return SBOM.from_json(infile.read())


@surfactant.plugin.hookimpl
def short_name() -> Optional[str]:
    return "cytrics"



import json
import subprocess
import time
from typing import List, Optional

import surfactant.plugin
from surfactant.plugin.manager import get_plugin_manager
from surfactant.sbomtypes import SBOM, Relationship, Software


@surfactant.plugin.hookimpl
def extract_file_info(
    sbom: SBOM, software: Software, filename: str, filetype: str, children: list
) -> Optional[List[Software]]:
    pm = get_plugin_manager()
    # Change to properly filter filetypes, add to if statement for filetypes syft should run for
    if filetype == "TAR":
        data = subprocess.check_output(
            "anchore_syft " + filename + " -o json --scope all-layers", shell=True
        )
        data = json.loads(data.decode())
        for i in data["artifacts"]:
            sw_entry = Software(
                sha1=None,
                # Syft does not provide a SHA256 for each artifact.
                # This uses their unique IDs in place since surfactant is dependent on this hash.
                # TODO (Ryan) discuss how to handle software entry without a hash with Lucas in the CyTRICS SBOM schema
                sha256=i["id"],
                md5=None,
                name=[i["name"]],
                fileName=None,
                installPath=[i["locations"][0]["path"]],
                containerPath=[filename],
                size=i["metadata"]["installedSize"],
                captureTime=int(time.time()),
                version=i["metadata"]["version"],
                vendor=[i["metadata"]["maintainer"]],
                description="",
                relationshipAssertion="Unknown",
                comments="Discovered using the Syft plugin with this cataloger: " + i["foundBy"],
                metadata=[],
                supplementaryFiles=[],
                provenance=None,
                components=[],
            )
            for file in i["metadata"]["files"]:
                if file["path"] not in sw_entry.supplementaryFiles:
                    sw_entry.supplementaryFiles.append(file["path"])
            for file in i["locations"]:
                if file["path"] not in sw_entry.supplementaryFiles:
                    sw_entry.supplementaryFiles.append(file["path"])
            children.append(sw_entry)
        gather_relationship_data(software, data, children)


@surfactant.plugin.hookimpl
def establish_relationships(
    sbom: SBOM, software: Software, metadata
) -> Optional[List[Relationship]]:
    relationship_list = []
    for meta in software.metadata:
        if "syftRelationships" in meta:
            for rel in meta["syftRelationships"]:
                relationship_list.append(Relationship(rel[0], rel[1], rel[2]))
    return relationship_list


def gather_relationship_data(image_sw: Software, data: str, sw_list: list):
    uuid_dict = {}
    # Build UUID dict for fast lookup
    # First entry is the image sw entry
    uuid_dict[data["source"]["id"]] = [-1, image_sw.UUID]
    for count, sw in enumerate(sw_list):
        index_uuid_list = [count, sw.UUID]
        uuid_dict[sw.sha256] = index_uuid_list
    for rel in data["artifactRelationships"]:
        # Both the parent and the child must have an existing software object
        if rel["parent"] in uuid_dict and rel["child"] in uuid_dict:
            parent_info = uuid_dict[rel["parent"]]
            child_info = uuid_dict[rel["child"]]
            if parent_info[0] == -1:
                sw = image_sw
            else:
                sw = sw_list[parent_info[0]]
            sw.relationshipAssertion = "Known"
            sw_list[child_info[0]].relationshipAssertion = "Known"
            relationship_list = []
            for meta in sw.metadata:
                if "syftRelationships" in meta:
                    relationship_list = meta["syftRelationships"]
                    break
            # Software object did not already have a list of syft relationships
            if len(relationship_list) == 0:
                sw.metadata.append({})
                sw.metadata[-1]["syftRelationships"] = relationship_list
            relationship_list.append([parent_info[1], child_info[1], rel["type"]])



import click

from surfactant.sbomtypes import SBOM


@click.command("stat")
@click.argument("input_sbom", type=click.File("r"), required=True)
def stat(input_sbom):
    data = SBOM.from_json(input_sbom.read())
    elfIsLib = 0
    elfIsExe = 0
    peIsExe = 0
    peIsdll = 0
    clrExe = 0
    clrDll = 0
    for sw in data.software:
        if not sw.metadata:
            continue
        for md in sw.metadata:
            if "elfIsLib" in md:
                if md["elfIsLib"]:
                    elfIsLib += 1
            if "elfIsExe" in md:
                if md["elfIsExe"]:
                    elfIsExe += 1
            if "peIsExe" in md:
                if md["peIsExe"]:
                    peIsExe += 1
            if "peIsDll" in md:
                if md["peIsDll"]:
                    peIsdll += 1
            if "peIsClr" in md:
                if md["peIsClr"]:
                    if md["peIsExe"]:
                        clrExe += 1
                    else:
                        clrDll += 1
    num_pe_exe_str = f"Number of PE Executables: {peIsExe} with {clrExe} using .NET/CLR"
    num_dll_str = f"Number of DLLs: {peIsdll} with {clrDll} using .NET/CLR"
    num_elf_bin_str = f"Number of ELF Binaries: {elfIsExe}"
    num_elf_shared_lib_str = f"Number of ELF shared libraries: {elfIsLib}"
    print(num_pe_exe_str)
    print(num_dll_str)
    print(num_elf_bin_str)
    print(num_elf_shared_lib_str)



# Copyright 2024 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import gzip
import subprocess
import tempfile
from typing import List, Optional

from loguru import logger

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


def check_if_grype_installed() -> bool:
    try:
        result = subprocess.run(["grype", "--help"], capture_output=True, check=False).returncode
    except FileNotFoundError:
        result = 1
    if result != 0:
        logger.warning("Install grype for the grype plugin to run")
    return result == 0


disable_plugin = not check_if_grype_installed()


def run_grype(filename: str) -> object:
    result = subprocess.run(["grype", filename], capture_output=True, check=False)
    if result.returncode != 0:
        logger.warning(f"Running grype on {filename} failed")
        return None
    output = result.stdout.decode()
    to_ret = []
    # skip the header on the first line
    for line in output.split("\n")[1:]:
        columns = [s.strip() for s in line.split("  ") if s.strip()]
        # Skip empty lines
        if len(columns) == 0:
            continue
        # Assume that the "Fixed In" field is missing if there's only 5 entries
        name = columns[0]
        installed = columns[1]
        if len(columns) == 5:
            fixed_in = ""
            type_ = columns[2]
            vuln = columns[3]
            severity = columns[4]
        else:
            fixed_in = columns[2]
            type_ = columns[3]
            vuln = columns[4]
            severity = columns[5]
        to_ret.append(
            {
                "name": name,
                "installed": installed,
                "fixed_in": fixed_in,
                "type": type_,
                "vulnerability": vuln,
                "severity": severity,
            }
        )
    return {"grype_output": to_ret}


@surfactant.plugin.hookimpl
def extract_file_info(
    sbom: SBOM, software: Software, filename: str, filetype: str, children: list
) -> Optional[List[Software]]:
    if disable_plugin or filetype not in ("DOCKER_TAR", "DOCKER_GZIP"):
        return None
    if filetype == "DOCKER_GZIP":
        with open(filename, "rb") as gzip_in:
            gzip_data = gzip_in.read()
        with tempfile.NamedTemporaryFile() as gzip_out:
            gzip_out.write(gzip.decompress(gzip_data))
            return run_grype(gzip_out.name)
    return run_grype(filename)


@surfactant.plugin.hookimpl
def update_db():
    # Example update logic
    try:
        subprocess.check_call(["grype", "db", "update"])
        return "Database updated successfully."
    except subprocess.CalledProcessError as e:
        return f"Failed to update database: {e}"


@surfactant.plugin.hookimpl
def short_name():
    return "grype"



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pathlib


def posix_normpath(path: str) -> pathlib.PurePosixPath:
    """Normalize a path to a POSIX path, with '..' path components removing the previous directory.
    This is similar to os.path.normpath, but it also removes leading '..' path components for relative
    paths. Note that symlinks are not followed, so in some cases the result may not be a valid path on
    the local filesystem. For example, posix_normpath("/a/b/../c") == "/a/c", but if "/a/b" is a symlink
    to "/x/y", then "/a/c" is not a valid path on the local filesystem. Future work could add an option
    to track symlinks encountered in order to resolve them."""
    posix_path = pathlib.PurePosixPath(path)

    # Remove '..' path component and preceding path component
    # PurePosixPath.parts is a tuple, so we can't modify it in-place
    parts = list(posix_path.parts)
    i = 0
    while i < len(parts):
        if parts[i] == "..":
            del parts[i]
            if i > 0:
                if i > 1 or parts[0] not in ("//", "/"):
                    del parts[i - 1]
                    i -= 1
        else:
            i += 1
    return pathlib.PurePosixPath(*parts)



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from ._analysisdata import AnalysisData
from ._file import File
from ._hardware import Hardware
from ._observation import Observation
from ._provenance import (
    AnalysisDataProvenance,
    HardwareProvenance,
    ObservationProvenance,
    SoftwareComponentProvenance,
    SoftwareProvenance,
    SystemProvenance,
)
from ._relationship import Relationship, StarRelationship
from ._sbom import SBOM
from ._software import Software, SoftwareComponent
from ._system import System

__all__ = [
    "File",
    "System",
    "Hardware",
    "Software",
    "SoftwareComponent",
    "AnalysisData",
    "Observation",
    "Relationship",
    "StarRelationship",
    "SystemProvenance",
    "HardwareProvenance",
    "SoftwareProvenance",
    "SoftwareComponentProvenance",
    "AnalysisDataProvenance",
    "ObservationProvenance",
    "SBOM",
]



import json
from pathlib import Path

import click
from loguru import logger


@click.command()
@click.argument("directory", type=click.Path(exists=True))
@click.option(
    "-o",
    "--output",
    type=str,
    help="Output JSON file name, defaults to end of directory string passed as input",
)
@click.option(
    "--install-prefix",
    type=str,
    default="/",
    help="Install prefix to use in the configuration (default: '/')",
)
def create_config(directory, output, install_prefix):
    """Create surfactant input configuration file based on input directory."""
    extract_paths = [Path(directory).as_posix()]
    config_dict = {"extractPaths": extract_paths, "installPrefix": install_prefix}
    config_out = [config_dict]
    output_file_name = output or Path(directory).stem + ".json"
    with open(output_file_name, "w") as json_file:
        json.dump(config_out, json_file, indent=4)
    logger.info(f"Data written to {output_file_name}")



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import pathlib
import re
from typing import Optional

from loguru import logger

import surfactant.plugin


@surfactant.plugin.hookimpl
def identify_file_type(filepath: str) -> Optional[str]:
    # pylint: disable=too-many-return-statements
    _filetype_extensions = {
        ".sh": "SHELL",
        ".bash": "BASH",
        ".zsh": "ZSH",
        ".py": "PYTHON",
        ".pyc": "PYTHON_COMPILED",
        ".js": "JAVASCRIPT",
        ".css": "CSS",
        ".html": "HTML",
        ".htm": "HTML",
        ".php": "PHP",
        ".bat": "BATCH",
        ".pl": "PERL_OR_PROLOG",
        ".pm": "PERL_MODULE",
    }
    _interpreters = {
        b"sh": "SHELL",
        b"bash": "BASH",
        b"zsh": "ZSH",
        b"php": "PHP",
        b"python": "PYTHON",
        b"python3": "PYTHON",
        b"perl": "PERL",
    }
    try:
        with open(filepath, "rb") as f:
            head = f.read(256)
            if head.startswith(b"<!DOCTYPE html>"):
                return "HTML"
            if head.startswith(b"#!") and b"\n" in head:
                end_line = head.index(b"\n")
                head = head[:end_line]
                for interpreter, filetype in _interpreters.items():
                    if re.search(interpreter, head):
                        return filetype
                return "SHEBANG"
    except FileNotFoundError:
        logger.warning(f"File not found: {filepath}")
        return None
    suffix = pathlib.Path(filepath).suffix.lower()
    if suffix in _filetype_extensions:
        return _filetype_extensions[suffix]
    return None



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import json
import os
import re
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Union

from loguru import logger

import surfactant.plugin
from surfactant.configmanager import ConfigManager
from surfactant.database_manager.database_utils import (
    calculate_hash,
    download_database,
    load_hash_and_timestamp,
    save_hash_and_timestamp,
)
from surfactant.sbomtypes import SBOM, Software

# Global configuration
DATABASE_URL = "https://raw.githubusercontent.com/e-m-b-a/emba/11d6c281189c3a14fc56f243859b0bccccce8b9a/config/bin_version_strings.cfg"


class NativeLibDatabaseManager:
    def __init__(self) -> None:
        self._native_lib_database: Optional[Dict[str, Any]] = None
        self.database_version_file_path = (
            ConfigManager().get_data_dir_path()
            / "infoextractors"
            / "native_lib_patterns"
            / "native_lib_patterns.toml"
        )
        self.pattern_key = "native_lib_patterns"
        self.pattern_file = "native_lib_patterns.json"
        self.source = "nativefile.emba"
        self.new_hash: Optional[str] = None
        self.download_timestamp: Optional[datetime] = None

    @property
    def native_lib_database(self) -> Optional[Dict[str, Any]]:
        if self._native_lib_database is None:
            self.load_db()
        return self._native_lib_database

    @property
    def pattern_info(self) -> Dict[str, Any]:
        return {
            "pattern_key": self.pattern_key,
            "pattern_file": self.pattern_file,
            "source": self.source,
            "hash_value": self.new_hash,
            "timestamp": self.download_timestamp,
        }

    def load_db(self) -> None:
        native_lib_file = (
            ConfigManager().get_data_dir_path()
            / "infoextractors"
            / "native_lib_patterns"
            / self.pattern_file
        )

        try:
            with open(native_lib_file, "r") as regex:
                self._native_lib_database = json.load(regex)
        except FileNotFoundError:
            logger.warning(
                "Native library pattern could not be loaded. Run `surfactant plugin update-db native_lib_patterns` to fetch the pattern database."
            )
            self._native_lib_database = None

    def get_database(self) -> Optional[Dict[str, Any]]:
        return self._native_lib_database


native_lib_manager = NativeLibDatabaseManager()


def supports_file(filetype: str) -> bool:
    return filetype in ("PE", "ELF", "MACHOFAT", "MACHOFAT64", "MACHO32", "MACHO64")


@surfactant.plugin.hookimpl
def extract_file_info(
    sbom: SBOM, software: Software, filename: str, filetype: str
) -> Optional[Dict[str, Any]]:
    if not supports_file(filetype):
        return None
    return extract_native_lib_info(filename)


def extract_native_lib_info(filename: str) -> Optional[Dict[str, Any]]:
    native_lib_info: Dict[str, Any] = {"nativeLibraries": []}
    native_lib_database = native_lib_manager.get_database()

    if native_lib_database is None:
        return None

    found_libraries: set = set()
    library_names: List[str] = []
    contains_library_names: List[str] = []

    base_filename = os.path.basename(filename)
    filenames_list = match_by_attribute("filename", base_filename, native_lib_database)
    if len(filenames_list) > 0:
        for match in filenames_list:
            library_name = match["isLibrary"]
            if library_name not in found_libraries:
                library_names.append(library_name)
                found_libraries.add(library_name)

    try:
        with open(filename, "rb") as native_file:
            filecontent = native_file.read()
        filecontent_list = match_by_attribute("filecontent", filecontent, native_lib_database)

        for match in filecontent_list:
            library_name = match["containsLibrary"]
            if library_name not in found_libraries:
                contains_library_names.append(library_name)
                found_libraries.add(library_name)

    except FileNotFoundError:
        logger.warning(f"File not found: {filename}")

    if library_names:
        native_lib_info["nativeLibraries"].append({"isLibrary": library_names})

    if contains_library_names:
        native_lib_info["nativeLibraries"].append({"containsLibrary": contains_library_names})

    return native_lib_info


def match_by_attribute(
    attribute: str, content: Union[str, bytes], patterns_database: Dict[str, Any]
) -> List[Dict[str, Any]]:
    libs: List[Dict[str, str]] = []
    for lib_name, lib_info in patterns_database.items():
        if attribute in lib_info:
            for pattern in lib_info[attribute]:
                if attribute == "filename":
                    if pattern.lower() == content.lower():
                        libs.append({"isLibrary": lib_name})

                elif attribute == "filecontent":
                    matches = re.search(pattern.encode("utf-8"), content)
                    if matches:
                        libs.append({"containsLibrary": lib_name})
    return libs


def parse_emba_cfg_file(content: str) -> Dict[str, Dict[str, List[str]]]:
    database: Dict[str, Dict[str, List[str]]] = {}
    lines = content.splitlines()
    filtered_lines: List[str] = []

    for line in lines:
        if not (line.startswith("#") or line.startswith("identifier")):
            filtered_lines.append(line)

    for line in filtered_lines:
        line = line.strip()

        fields = line.split(";")

        lib_name = fields[0]

        name_patterns: List[str] = []

        if fields[3].startswith('"') and fields[3].endswith('""'):
            filecontent = fields[3][1:-1]
        elif fields[3].endswith('""'):
            filecontent = fields[3][:-1]
        else:
            filecontent = fields[3].strip('"')

        if fields[1] == "" or fields[1] == "strict":
            if fields[1] == "strict":
                if lib_name not in database:
                    database[lib_name] = {
                        "filename": [lib_name],
                        "filecontent": [],
                    }
                else:
                    if lib_name not in database[lib_name]["filename"]:
                        database[lib_name]["filename"].append(lib_name)
            else:
                try:
                    re.search(filecontent.encode("utf-8"), b"")
                    if lib_name not in database:
                        database[lib_name] = {
                            "filename": name_patterns,
                            "filecontent": [filecontent],
                        }
                    else:
                        database[lib_name]["filecontent"].append(filecontent)
                except re.error as e:
                    logger.error(f"Error parsing file content regexp {filecontent}: {e}")

    return database


@surfactant.plugin.hookimpl
def update_db() -> str:
    file_content = download_database(DATABASE_URL)
    if file_content is not None:
        native_lib_manager.new_hash = calculate_hash(file_content)
        current_data = load_hash_and_timestamp(
            native_lib_manager.database_version_file_path,
            native_lib_manager.pattern_key,
            native_lib_manager.pattern_file,
        )
        if current_data and native_lib_manager.new_hash == current_data.get("hash"):
            return "No update occurred. Database is up-to-date."

        parsed_data = parse_emba_cfg_file(file_content)
        for _, value in parsed_data.items():
            filecontent_list = value["filecontent"]

            for i, pattern in enumerate(filecontent_list):
                if pattern.startswith("^"):
                    filecontent_list[i] = pattern[1:]

                if not pattern.endswith("\\$"):
                    if pattern.endswith("$"):
                        filecontent_list[i] = pattern[:-1]

        path = ConfigManager().get_data_dir_path() / "infoextractors" / "native_lib_patterns"
        path.mkdir(parents=True, exist_ok=True)
        native_lib_file = path / native_lib_manager.pattern_file
        with open(native_lib_file, "w") as json_file:
            json.dump(parsed_data, json_file, indent=4)

        native_lib_manager.download_timestamp = datetime.now(timezone.utc)
        save_hash_and_timestamp(
            native_lib_manager.database_version_file_path, native_lib_manager.pattern_info
        )
        return "Update complete."
    return "No update occurred."


@surfactant.plugin.hookimpl
def short_name() -> Optional[str]:
    return "native_lib_patterns"


@surfactant.plugin.hookimpl
def init_hook(command_name: Optional[str] = None) -> None:
    """
    Initialization hook to load the native library database.

    Args:
        command_name (Optional[str], optional): The name of the command invoking the initialization.
            If set to "update-db", the database will not be loaded.

    Returns:
        None
    """
    if command_name != "update-db":
        logger.info("Initializing native_lib_file...")
        native_lib_manager.load_db()
        logger.info("Initializing native_lib_file complete.")



import pathlib
from collections.abc import Iterable
from typing import Dict, List, Optional, Tuple

import cyclonedx.output
from cyclonedx.model import HashAlgorithm, HashType
from cyclonedx.model.bom import Bom, BomMetaData
from cyclonedx.model.bom_ref import BomRef
from cyclonedx.model.component import Component, ComponentType
from cyclonedx.model.contact import OrganizationalEntity
from cyclonedx.model.dependency import Dependency
from cyclonedx.model.tool import Tool

import surfactant.plugin
from surfactant import __version__ as surfactant_version
from surfactant.sbomtypes import SBOM, Software, System


@surfactant.plugin.hookimpl
def write_sbom(sbom: SBOM, outfile) -> None:
    """Writes the contents of the SBOM to a CycloneDX file.

    The write_sbom hook for the cyclonedx_writer makes a best-effort attempt
    to map the information gathered from the internal SBOM representation
    to a valid CycloneDX file.

    Args:
        sbom (SBOM): The SBOM to write to the output file.
        outfile: The output file handle to write the SBOM to.
    """
    # NOTE eventually outformat and some fields in CycloneDX document should be user settable
    outformat = "json"

    # or Bom(tools=[surfactant_tool])
    surfactant_tool = Tool(name="Surfactant", version=f"{surfactant_version}")
    bom_metadata = BomMetaData(tools=[surfactant_tool])
    bom = Bom(metadata=bom_metadata)

    # Add CycloneDX Components for systems
    for system in sbom.systems:
        system_uuid, comp = convert_system_to_cyclonedx_component(system)
        bom.components.add(comp)

    # Create a map of CycloneDX IDs for Components that map directly to a known container UUID
    # this gets used to avoid creating excessive relationships if a file is present
    # due to being found within multiple different containers
    container_path_relationships: Dict[Tuple[str, str]] = {}

    for software in sbom.software:
        # Create CycloneDX Components for every software entry
        # start with software entries that act as containers for other software entries
        if sbom.has_relationship(xUUID=software.UUID, relationship="Contains"):
            _, container_list = convert_software_to_cyclonedx_container_components(software)
            for container in container_list:
                bom.components.add(container)
        else:
            for parent_uuid, _, file in convert_software_to_cyclonedx_file_components(software):
                bom.components.add(file)
                if parent_uuid:
                    container_path_relationships[file.bom_ref.value] = parent_uuid

    # Convert relationships into CycloneDX relationships
    cdx_rels: Dict[str, Dependency] = {}
    for rel in sbom.relationships:
        rel_type = rel.relationship.upper()

        # Minimize duplicate contains relationships for files with multiple container paths
        if (
            (rel_type == "CONTAINS")
            and (rel.yUUID in container_path_relationships)
            and (rel.xUUID != container_path_relationships[rel.yUUID])
        ):
            continue

        # Create Dependency instance for relationships where xUUID is the parent
        if rel.xUUID not in cdx_rels:
            cdx_rels[rel.xUUID] = Dependency(ref=BomRef(rel.xUUID))
        rel_dependency = cdx_rels[rel.xUUID]

        if rel.yUUID not in rel_dependency.dependencies:
            rel_dependency.dependencies.add(Dependency(ref=BomRef(rel.yUUID)))

    # Add dependencies for each component
    for _, v in cdx_rels.items():
        bom.dependencies.add(v)

    # Write the CycloneDX SBOM file
    output_format = output_format = cyclonedx.output.OutputFormat.JSON
    if outformat == "json":
        output_format = cyclonedx.output.OutputFormat.JSON
    elif outformat == "xml":
        output_format = cyclonedx.output.OutputFormat.XML
    # The docs say that you don't need to specify a version (it says it defaults to the latest)
    # but I got a missing keyword error when doing so, so just specify 1.5 for now
    outputter: cyclonedx.output.BaseOutput = cyclonedx.output.make_outputter(
        bom=bom, output_format=output_format, schema_version=cyclonedx.schema.SchemaVersion.V1_5
    )
    outfile.write(outputter.output_as_string())


@surfactant.plugin.hookimpl
def short_name() -> Optional[str]:
    return "cyclonedx"


def convert_system_to_cyclonedx_component(system: System) -> Tuple[str, Component]:
    """Converts a system entry in the SBOM to a CycloneDX Component.

    If a system entry has multiple vendors, only the first one is chosen as the
    supplier for the CycloneDX Component.

    Args:
        system (System): The SBOM system to convert to a CycloneDX Component.

    Returns:
        Tuple[str, Component]: A tuple containing the UUID of the system that was
        converted into a Component, and the CycloneDX Component object that was created.
    """
    # Pick the best name for the package
    name = system.officialName
    if not name and system.name:
        name = system.name

    # Pick a vendor to use as the supplier
    supplier = None
    if system.vendor:
        # assume organization, can only list one supplier
        supplier = OrganizationalEntity(name=system.vendor[0])

    system_component = Component(
        bom_ref=system.UUID,
        name=name,
        supplier=supplier,
        description=system.description,
        # components: Optional[Iterable['Component']]
        type=ComponentType.CONTAINER,
    )

    return system.UUID, system_component


def convert_software_to_cyclonedx_container_components(
    software: Software,
) -> Tuple[str, List[Component]]:
    """Converts a software entry in the SBOM to one or more CycloneDX Components.

    A CycloneDX Component is created for each file name that the software can have. If
    a software entry has multiple vendors, only the first one is chosen as the
    supplier for the CycloneDX Component.

    Args:
        software (Software): The SBOM software entry to convert to CycloneDX Components.

    Returns:
        Tuple[str, List[Component]]: A tuple containing the UUID of the software that was
        converted into Components, and a list of the CycloneDX Component objects that were created.
    """
    containers: List[Component] = []
    for fname in software.fileName:
        name = software.name
        if not name:
            # No name, fall-back to the file name
            name = fname
        # Pick a vendor to use as the supplier
        supplier = None
        if software.vendor:
            # assume organization, can only list one supplier
            supplier = OrganizationalEntity(name=software.vendor[0])
        hashes = []
        if software.sha1:
            hashes.append(HashType(alg=HashAlgorithm.SHA_1, content=software.sha1))
        if software.sha256:
            hashes.append(HashType(alg=HashAlgorithm.SHA_256, content=software.sha256))
        if software.md5:
            hashes.append(HashType(alg=HashAlgorithm.MD5, content=software.md5))
        if not hashes:
            hashes = None
        containers.append(
            Component(
                bom_ref=software.UUID,
                name=name,
                version=software.version,
                supplier=supplier,
                description=software.description,
                hashes=hashes,
                # components: Optional[Iterable['Component']]
                type=ComponentType.CONTAINER,
            )
        )
    return software.UUID, containers


def convert_software_to_cyclonedx_file_components(
    software: Software,
) -> List[Tuple[str, str, Component]]:
    """Converts a software entry in the SBOM to one or more CycloneDX FILE Components.

    A CycloneDX Component is created for each unique container path that the software has. If
    no container paths exist, each unique file name will be used instead. If a software
    entry has multiple vendors, only the first one is chosen as the supplier for the
    CycloneDX Component.

    Args:
        software (Software): The SBOM software entry to convert to CycloneDX Components.

    Returns:
        List[Tuple[str, str, Component]]: A list of tuples that contains the UUID of the parent
        container for the software entry (or None if file names were used), the UUID of the
        software entry that was converted into a CycloneDX Component, and the resulting CycloneDX
        Component that was created.
    """
    files: List[Tuple[str, str, Component]] = []
    for cpathstr in software.containerPath:
        cpath = pathlib.PurePath(cpathstr)
        # Less than 2 parts would just be the container path uuid, or a file name
        if len(cpath.parts) > 1:
            # First entry in container path is the parent container UUID
            parent_uuid = cpath.parts[0]
            # Full path to file, relative to package root (starting with "./")
            file_path = "/".join(cpath.parts[1:])

            file = create_cyclonedx_file(file_path, software)
            files.append((parent_uuid, software.UUID, file))
    # Alternative if no container paths exist for a software entry
    if not software.containerPath:
        for fname in software.fileName:
            file = create_cyclonedx_file(fname, software)
            files.append((None, software.UUID, file))
    return files


def create_cyclonedx_file(file_path: str, software: Software) -> Component:
    """Creates a CycloneDX FILE Component from a software entry.

    Args:
        file_path (str): The path relative to the parent container for the CycloneDX FILE Component.
        software (Software): The SBOM software entry to convert to a CycloneDX FILE Component.

    Returns:
        Component: CycloneDX FILE Component with information filled in based on the provided software entry.
    """
    # Pick a vendor to use as the supplier
    supplier = None
    if software.vendor:
        # assume organization, can only list one supplier
        supplier = OrganizationalEntity(name=software.vendor[0])

    hashes = []
    if software.sha1:
        hashes.append(HashType(alg=HashAlgorithm.SHA_1, content=software.sha1))
    if software.sha256:
        hashes.append(HashType(alg=HashAlgorithm.SHA_256, content=software.sha256))
    if software.md5:
        hashes.append(HashType(alg=HashAlgorithm.MD5, content=software.md5))
    if not hashes:
        hashes = None

    copyright_text = None
    if cr_text := get_fileinfo_metadata(software, "LegalCopyright"):
        copyright_text = cr_text  # free-form text field extracted from actual file identifying copyright holder and any dates present

    if software.description == "":
        software.description = None

    if software.version == "":
        software.version = None

    return Component(
        bom_ref=software.UUID,
        name=file_path,
        version=software.version,
        supplier=supplier,
        description=software.description,
        hashes=hashes,
        copyright=copyright_text,
        # components: Optional[Iterable['Component']]
        type=ComponentType.FILE,
    )


def get_fileinfo_metadata(software: Software, field: str) -> Optional[str]:
    """Retrieves the value for a field in a 'FileInfo' metadata object in a software entry.

    Args:
        software (Software): The software entry to get the 'FileInfo' field from.
        field (str): The name of the 'FileInfo' field to get the value of.

    Returns:
        Optional[str]: The value of the 'FileInfo' metadata field request, or None.
    """
    if software.metadata and isinstance(software.metadata, Iterable):
        for entry in software.metadata:
            if "FileInfo" in entry and field in entry["FileInfo"]:
                return entry["FileInfo"][field]
    return None


def get_software_field(software: Software, field: str):
    """Retrieves the value for a field in a SBOM software entry.

    The field retrieved can be the name of an attribute in the Software dataclass,
    or if the field provided is 'Copyright' it will try to retrieve the copyright
    information from a 'FileInfo' metadata object.

    Args:
        software (Software): The software entry to read the field from.
        field (str): The name of the field to get the value of, or 'Copyright'.

    Returns:
        Optional[str]: The value of the requested field, or None.
    """
    if hasattr(software, field):
        return getattr(software, field)
    # Copyright field currently only gets populated from Windows PE file metadata
    if field == "Copyright":
        if software.metadata and isinstance(software.metadata, Iterable):
            for entry in software.metadata:
                if "FileInfo" in entry and "LegalCopyright" in entry["FileInfo"]:
                    return entry["FileInfo"]["LegalCopyright"]
    return None



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import hashlib
import pathlib
import random
import string
import sys
import uuid
from collections.abc import Iterable
from datetime import datetime
from typing import Dict, List, Optional, Tuple

import spdx_tools.spdx.writer.json.json_writer as jsonwriter
import spdx_tools.spdx.writer.tagvalue.tagvalue_writer as tvwriter
from spdx_tools.spdx.model.actor import Actor, ActorType
from spdx_tools.spdx.model.checksum import Checksum, ChecksumAlgorithm
from spdx_tools.spdx.model.document import CreationInfo, Document
from spdx_tools.spdx.model.file import File, FileType
from spdx_tools.spdx.model.package import Package
from spdx_tools.spdx.model.relationship import Relationship, RelationshipType
from spdx_tools.spdx.model.spdx_no_assertion import SpdxNoAssertion
from spdx_tools.spdx.spdx_element_utils import calculate_package_verification_code

import surfactant.plugin
from surfactant import __version__ as surfactant_version
from surfactant.sbomtypes import SBOM, Software, System


@surfactant.plugin.hookimpl
def write_sbom(sbom: SBOM, outfile) -> None:
    """Writes the contents of the SBOM to a SPDX file.

    The write_sbom hook for the spdx_writer makes a best-effort attempt
    to map the information gathered from the internal SBOM representation
    to a valid SPDX file. Currently the spdx-tools library used for validating
    and writing a SPDX file has an unstable API which can at times lead
    to things not working.

    Args:
        sbom (SBOM): The SBOM to write to the output file.
        outfile: The output file handle to write the SBOM to.
    """
    # NOTE eventually outformat and many fields in SPDX document should be user settable (namespace, name)
    outformat = "json"

    spdx_doc = create_spdx_doc()

    # NOTE CyTRICS SBOM format lumps all file names/container paths together into a single list; preserving
    # a semblance of a file tree with nested filesystem/archive information would help some for establishing
    # SPDX Relationships -- though in general SBOM format conversion isn't straightforward or lossless

    # Create a map of UUIDs to SPDX IDs, for defining SPDX Relationships later
    uuid_to_spdxid: Dict[str, List[str]] = {}

    # Add SPDX packages for systems
    for system in sbom.systems:
        system_uuid, pkg = convert_system_to_spdx_package(system)
        spdx_doc.packages = spdx_doc.packages + [pkg]
        if system_uuid not in uuid_to_spdxid:
            uuid_to_spdxid[system_uuid] = []
        uuid_to_spdxid[system_uuid].append(pkg.spdx_id)

    # Create a map of SPDX IDs for Files that map directly to a known container UUID
    # this gets used to avoid creating excessive relationships if a file is present
    # due to being found within multiple different containers
    container_path_relationships: Dict[Tuple[str, str]] = {}

    for software in sbom.software:
        # Add software entries that contain others as SPDX Packages
        # anything containing other files can be a SPDX Package (directory, container, tarball, zip, etc)
        # some could also be represented as a SPDX File
        if sbom.has_relationship(xUUID=software.UUID, relationship="Contains"):
            software_uuid, pkgs = convert_software_to_spdx_packages(software)
            for pkg in pkgs:
                spdx_doc.packages = spdx_doc.packages + [pkg]
                if software_uuid not in uuid_to_spdxid:
                    uuid_to_spdxid[software_uuid] = []
                uuid_to_spdxid[software_uuid].append(pkg.spdx_id)
        # Add all other software entries as SPDX Files
        else:
            for parent_uuid, software_uuid, file in convert_software_to_spdx_files(software):
                spdx_doc.files = spdx_doc.files + [file]
                if parent_uuid:
                    container_path_relationships[file.spdx_id] = parent_uuid
                if software_uuid:
                    if software_uuid not in uuid_to_spdxid:
                        uuid_to_spdxid[software_uuid] = []
                    uuid_to_spdxid[software_uuid].append(file.spdx_id)

    # Add describes relationship between SPDX Document and Packages
    for pkg in spdx_doc.packages:
        spdx_rel = Relationship(
            spdx_element_id=spdx_doc.creation_info.spdx_id,
            relationship_type=RelationshipType.DESCRIBES,
            related_spdx_element_id=pkg.spdx_id,
        )
        spdx_doc.relationships = spdx_doc.relationships + [spdx_rel]

    # No packages, don't make any assertions about what the SPDX Document describes
    if not spdx_doc.relationships:
        spdx_rel = Relationship(
            spdx_element_id=spdx_doc.creation_info.spdx_id,
            relationship_type=RelationshipType.DESCRIBES,
            related_spdx_element_id=SpdxNoAssertion(),
        )
        spdx_doc.relationships = spdx_doc.relationships + [spdx_rel]

    # Convert relationships into SPDX Relationships
    for rel in sbom.relationships:
        if (rel.xUUID in uuid_to_spdxid) and (rel.yUUID in uuid_to_spdxid):
            # UUID to SPDX ID is (potentially) one-to-many mapping
            for x_spdxid in uuid_to_spdxid[rel.xUUID]:
                for y_spdxid in uuid_to_spdxid[rel.yUUID]:
                    rel_type = rel.relationship.upper()  # Contains will map to SPDX "CONTAINS"

                    # Minimize duplicate contains relationships for files with multiple container paths
                    if (
                        (rel_type == "CONTAINS")
                        and (y_spdxid in container_path_relationships)
                        and (rel.xUUID != container_path_relationships[y_spdxid])
                    ):
                        continue

                    rel_comment = None  # Default value is None
                    try:
                        # Check if rel_type is defined by SPDX; otherwise throws KeyError
                        RelationshipType[rel_type]
                    except KeyError:
                        # Use type "OTHER" and a comment if SPDX doesn't define the relationship type
                        rel_comment = f"Type: {rel_type}"
                        rel_type = "OTHER"
                    spdx_rel = Relationship(
                        spdx_element_id=x_spdxid,
                        relationship_type=RelationshipType[rel_type],
                        related_spdx_element_id=y_spdxid,
                        comment=rel_comment,
                    )
                    spdx_doc.relationships = spdx_doc.relationships + [spdx_rel]

    # Add package verification codes
    for pkg in spdx_doc.packages:
        files = []
        for file in spdx_doc.files:
            pkg_contains_relationships = {
                relationship.related_spdx_element
                for relationship in spdx_doc.relationships
                if relationship.relationship_type == "CONTAINS"
                and relationship.spdx_element_id == pkg.spdx_id
            }
            pkg_contained_by_relationships = {
                relationship.spdx_element_id
                for relationship in spdx_doc.relationships
                if relationship.relationship_type == "CONTAINED_BY"
                and relationship.related_spdx_element == pkg.spdx_id
            }
            if file.spdx_id in pkg_contains_relationships | pkg_contained_by_relationships:
                files.append(file)
        pkg.verification_code = calculate_package_verification_code(files)

    if outformat == "json":
        try:
            jsonwriter.write_document_to_stream(spdx_doc, outfile)
        except ValueError as e:
            sys.stderr.write(str(e))
    elif outformat == "tagvalue":
        try:
            tvwriter.write_document_to_stream(spdx_doc, outfile)
        except ValueError as e:
            sys.stderr.write(str(e))


@surfactant.plugin.hookimpl
def short_name() -> Optional[str]:
    return "spdx"


def convert_system_to_spdx_package(system: System) -> Tuple[str, Package]:
    """Converts a system entry in the SBOM to a SPDX Package.

    If a system entry has multiple vendors, only the first one is chosen as the
    supplier for the SPDX Package (due to limitations of the SPDX format).

    Args:
        system (System): The SBOM system to convert to a SPDX Package.

    Returns:
        Tuple[str, Package]: A tuple containing the UUID of the system that was
        converted into a Package, and the SPDX Package object that was created.
    """
    # Pick the best name for the package
    name = system.officialName
    if not name and system.name:
        name = system.name

    # Pick a vendor to use as the supplier
    supplier = None
    if system.vendor:
        # assume Organization, not enough info to distinguish People
        supplier = Actor(name=system.vendor[0], actor_type=ActorType.ORGANIZATION)

    return system.UUID, create_spdx_package(name, system.description, supplier)


def convert_software_to_spdx_packages(software: Software) -> Tuple[str, List[Package]]:
    """Converts a software entry in the SBOM to one or more SPDX Packages.

    A SPDX Package is created for each file name that the software can have. If
    a software entry has multiple vendors, only the first one is chosen as the
    supplier for the SPDX Package (due to limitations of the SPDX format).

    Args:
        software (Software): The SBOM software entry to convert to SPDX Packages.

    Returns:
        Tuple[str, List[Package]]: A tuple containing the UUID of the software that was
        converted into Packages, and a list of the SPDX Package objects that were created.
    """
    packages: List[Package] = []
    for fname in software.fileName:
        name = software.name
        if not name:
            # No name, fall-back to the file name
            name = fname
        # Pick a vendor to use as the supplier
        supplier = None
        if software.vendor:
            # assume Organization, not enough info to distinguish People
            supplier = Actor(name=software.vendor[0], actor_type=ActorType.ORGANIZATION)
        packages.append(
            create_spdx_package(
                name,
                software.description,
                supplier,
                file_name=fname,
                version=software.version,
                sha1=software.sha1,
                sha256=software.sha256,
                md5=software.md5,
            )
        )
    return software.UUID, packages


def convert_software_to_spdx_files(software: Software) -> List[Tuple[str, str, File]]:
    """Converts a software entry in the SBOM to one or more SPDX Files.

    A SPDX File is created for each unique container path that the software has. If
    no container paths exist, each unique file name will be used instead. If a software
    entry has multiple vendors, only the first one is chosen as the supplier for the SPDX
    File (due to limitations of the SPDX format).

    Args:
        software (Software): The SBOM software entry to convert to SPDX Files.

    Returns:
        List[Tuple[str, str, File]]: A list of tuples that contains the UUID of the parent
        container for the software entry (or None if file names were used), the UUID of the
        software entry that was converted into a SPDX File, and the resulting SPDX File that
        was created.
    """
    files: List[Tuple[str, str, File]] = []
    for cpathstr in software.containerPath:
        cpath = pathlib.PurePath(cpathstr)
        # Less than 2 parts would just be the container path uuid, or a file name
        if len(cpath.parts) > 1:
            # First entry in container path is the parent container UUID
            parent_uuid = cpath.parts[0]
            # Full path to file, relative to package root (starting with "./")
            file_path = "./" + "/".join(cpath.parts[1:])

            idstring = generate_file_idstring(software, cpath.parts[-1])
            file = create_spdx_file(idstring, file_path, software)
            files.append((parent_uuid, software.UUID, file))
    # Alternative if no container paths exist for a software entry
    if not software.containerPath:
        for fname in software.fileName:
            idstring = generate_file_idstring(software, fname)
            file = create_spdx_file(idstring, "./" + fname, software)
            files.append((None, software.UUID, file))
    return files


def create_spdx_doc() -> Document:
    """Creates a SPDX Document with some default values for required fields filled in.

    Returns:
        Document: The SPDX Document that was created.
    """
    # NOTE: The document namespace URI does not have to be accessible, it just provides a unique ID.
    # Format is: https://[CreatorWebsite]/[pathToSpdx]/[DocumentName]-[UUID]
    # spdx.org/spdxdocs can be used if the creator does not own their own website.
    doc_name = "SBOM - DRAFT"  # document name, as designated by the creator
    spdx_creationinfo = CreationInfo(
        spdx_version="SPDX-2.3",
        spdx_id="SPDXRef-DOCUMENT",
        name=doc_name,
        document_namespace=f"https://spdx.org/spdxdocs/{doc_name}-{str(uuid.uuid4())}",  # unique URI for this SPDX Document
        creators=[
            Actor(name=f"Surfactant-{surfactant_version}", actor_type=ActorType.TOOL)
        ],  # Organization or Person can also be added as creators (may use "anonymous")
        created=datetime.utcnow().replace(microsecond=0),
        creator_comment="This SPDX document was created by using Surfactant.",
        document_comment="This is a DRAFT SPDX document, and is incomplete.",
    )
    spdx_doc = Document(creation_info=spdx_creationinfo)
    return spdx_doc


def create_spdx_file(idstring: str, file_path: str, software: Software) -> File:
    """Creates a SPDX File from a software entry.

    At minimum, the software entry must have a valid 'sha1' checksum.

    Args:
        idstring (str): A unique SPDX ID string (consisting of only alphanumeric, '.', and '-' characters).
        file_path (str): The full path to the SPDX File (starting with './', relative to the package/container).
        software (Software): The SBOM software entry to convert to a SPDX File.

    Returns:
        File: A SPDX File with information filled in based on the provided software entry.
    """
    sw_checksums = [
        Checksum(ChecksumAlgorithm.SHA1, software.sha1.lower())
    ]  # SHA1 required, should probably error if doesn't exist
    if software.sha256:
        sw_checksums.append(Checksum(ChecksumAlgorithm.SHA256, software.sha256.lower()))
    if software.md5:
        sw_checksums.append(Checksum(ChecksumAlgorithm.MD5, software.md5.lower()))

    sw_copyright = (
        SpdxNoAssertion()
    )  # SPDXNone for nothing present, SpdxNoAssertion for no attempt to determine
    if cr_text := get_fileinfo_metadata(software, "LegalCopyright"):
        sw_copyright = cr_text  # free-form text field extracted from actual file identifying copyright holder and any dates present

    file = File(
        name=file_path,
        spdx_id=f"SPDXRef-{idstring}",
        checksums=sw_checksums,
        file_types=[
            FileType.BINARY
        ],  # One or more of: SOURCE | BINARY | ARCHIVE | APPLICATION | AUDIO | IMAGE | TEXT | VIDEO | DOCUMENTATION | SPDX | OTHER
        license_concluded=SpdxNoAssertion(),  # SPDXNone if no license available for file, SpdxNoAssertion if can't determine
        license_info_in_file=[
            SpdxNoAssertion()
        ],  # info in actual file (e.g. header, not external such as COPYING.txt); None for nothing, SpdxNoAssertion for did not look
        copyright_text=sw_copyright,
    )

    return file


def create_spdx_package(
    name: str,
    summary,
    supplier,
    *,  # all remaining arguments are keyword-only
    file_name: Optional[str] = None,
    version: Optional[str] = None,
    sha1: Optional[str] = None,
    sha256: Optional[str] = None,
    md5: Optional[str] = None,
) -> Package:
    """Creates a SPDX Package from the provided information.

    If there is a checksum provided, at minimum the 'sha1' should be present.

    Args:
        name (str): Name of the SPDX package.
        summary: A concise summary of the function or use of the package.
        supplier: The vendor the package came from.
        file_name (Optional[str]): Actual file name or path to the directory that is being treated as a package (subdirectory is denoted with './').
        version (Optional[str]): Version identifier for the package.
        sha1 (Optional[str]): SHA1 checksum that uniquely identifies the package. If the package includes the SPDX file, this should NOT be present.
        sha256 (Optional[str]): SHA256 checksum that uniquely identifies the package. If the package includes the SPDX file, this should NOT be present.
        md5 (Optional[str]): MD5 checksum that uniquely identifies the package. If the package includes the SPDX file, this should NOT be present.

    Returns:
        Package: A SPDX Package with information filled in based on the provided information.
    """
    pkg_checksums = []
    if sha1:
        pkg_checksums.append(Checksum(ChecksumAlgorithm.SHA1, sha1.lower()))
    if sha256:
        pkg_checksums.append(Checksum(ChecksumAlgorithm.SHA256, sha256.lower()))
    if md5:
        pkg_checksums.append(Checksum(ChecksumAlgorithm.MD5, md5.lower()))

    idstring = generate_package_idstring(name, version, file_name)
    pkg = Package(
        spdx_id=f"SPDXRef-{idstring}",
        name=name,
        download_location=SpdxNoAssertion(),  # SPDXNone if there is no location whatsoever, not just that we "failed" or didn't try to locate one
        license_concluded=SpdxNoAssertion(),  # if different from declared license, must explain in comments why; SpdxNoAssertion prefer to have comment
        license_declared=SpdxNoAssertion(),  # licenses declared by authors of package, not 3rd party repo; SpdxNoAssertion since didn't attempt to determine
        license_info_from_files=[
            SpdxNoAssertion()
        ],  # SPDXNone if no license info detected, SpdxNoAssertion if didn't try to determine
        copyright_text=SpdxNoAssertion(),  # SPDXNone if no copyright info, SpdxNoAssertion if didn't try to determine; use any Cr text even if incomplete
        version=version if version else None,
        file_name=(
            file_name if file_name else None
        ),  # actual file name, or path to directory treated as package (subdirectory is denoted with ./)
        supplier=(
            supplier if supplier else SpdxNoAssertion()
        ),  # SpdxNoAssertion if can't determine (SPDX can't handle multiple vendors)
        originator=SpdxNoAssertion(),  # 3rd party who distributed package is different than the supplier/vendor
        checksums=pkg_checksums,
        summary=(
            summary if summary else None
        ),  # concise info on the function or use of package, without having to parse source code
        homepage=SpdxNoAssertion(),  # SPDXNone if none exists, SpdxNoAssertion if didn't try to find a homepage
    )
    # primary_package_purpose can be: APPLICATION | FRAMEWORK | LIBRARY | CONTAINER | OPERATING - SYSTEM |
    # DEVICE | FIRMWARE | SOURCE | ARCHIVE | FILE | INSTALL | OTHER

    return pkg


def generate_random_idstring(num_chars: int = 5) -> str:
    """Generate a unique random alphanumeric ID string.

    Args:
        num_chars (int): Number of characters (default=5) that should be in the generated id string.

    Returns:
        str: The randomly generated idstring.
    """
    return "".join(random.choices(string.ascii_letters + string.digits, k=num_chars))


def generate_file_idstring(software: Software, filename: str) -> str:
    """Generate a unique random alphanumeric ID string for a SPDX File.

    The generated id string will consist of the provided file name, software product name,
    software version, and a randomly generated string separated by '-'. Non-alphanumeric or
    '.' or '-' characters will be removed from those components, and empty components will
    be omitted.

    Args:
        software (Software): Software entry to get product name and version to use in the id string.
        filename (str): File name (one of possibly several in software) to include in the id string.

    Returns:
        str: The randomly generated idstring for a SPDX File.
    """
    # Filename
    idfilename = "".join(ch for ch in filename if (ch.isalnum() or ch in ".-"))
    # Product name
    idproductname = ""
    if software.name:
        idproductname = "".join(ch for ch in software.name if (ch.isalnum() or ch in ".-"))
    # Version
    idversion = ""
    if software.version:
        idversion = "".join(ch for ch in software.version if (ch.isalnum() or ch in ".-"))
    # Return id string with generated unique id appended to ensure no duplicate SPDXRefs for files
    # leaving out "fields" that are empty strings
    return "-".join(
        x for x in [idfilename, idversion, idproductname, generate_random_idstring()] if x
    )


def generate_package_idstring(name: str, version: str, file_name: str) -> str:
    """Generate a unique random alphanumeric ID string for a SPDX Package.

    The generated id string will consist of the provided name, file name, version, and a
    randomly generated string separated by '-'. Non-alphanumeric or '.' or '-' characters
    will be removed from those components, and empty components will be omitted.

    Args:
        name (str): Name to include in the id string.
        version (str): Version to include in the id string.
        file_name (str): File name to include in the id string.

    Returns:
        str: The randomly generated idstring for a SPDX Package.
    """
    # Package Name
    idname = ""
    if idname:
        idname = "".join(ch for ch in name if (ch.isalnum() or ch in ".-"))
    # File Name
    idfilename = ""
    if file_name:
        idfilename = "".join(ch for ch in file_name if (ch.isalnum() or ch in ".-"))
    # Package Version
    idversion = ""
    if version:
        idversion = "".join(ch for ch in version if (ch.isalnum() or ch in ".-"))
    # Return id string with generated unique id appended to ensure no duplicate SPDXRefs for files
    # leaving out "fields" that are empty strings
    return "-".join(x for x in [idname, idversion, idfilename, generate_random_idstring()] if x)


def get_fileinfo_metadata(software: Software, field: str) -> Optional[str]:
    """Retrieves the value for a field in a 'FileInfo' metadata object in a software entry.

    Args:
        software (Software): The software entry to get the 'FileInfo' field from.
        field (str): The name of the 'FileInfo' field to get the value of.

    Returns:
        Optional[str]: The value of the 'FileInfo' metadata field request, or None.
    """
    if software.metadata and isinstance(software.metadata, Iterable):
        for entry in software.metadata:
            if "FileInfo" in entry and field in entry["FileInfo"]:
                return entry["FileInfo"][field]
    return None


def get_software_field(software: Software, field: str):
    """Retrieves the value for a field in a SBOM software entry.

    The field retrieved can be the name of an attribute in the Software dataclass,
    or if the field provided is 'Copyright' it will try to retrieve the copyright
    information from a 'FileInfo' metadata object.

    Args:
        software (Software): The software entry to read the field from.
        field (str): The name of the field to get the value of, or 'Copyright'.

    Returns:
        Optional[str]: The value of the requested field, or None.
    """
    if hasattr(software, field):
        return getattr(software, field)
    # Copyright field currently only gets populated from Windows PE file metadata
    if field == "Copyright":
        if software.metadata and isinstance(software.metadata, Iterable):
            for entry in software.metadata:
                if "FileInfo" in entry and "LegalCopyright" in entry["FileInfo"]:
                    return entry["FileInfo"]["LegalCopyright"]
    return None


def java_generate_package_verification_code(software: List[Software]) -> Tuple[str, List[str]]:
    """Generate a SPDX package verif_code according to the method used by the Java SPDX Tools.

    This is not the algorithm defined in the SPDX specification. The implementation here is provided
    in case we need to support an SBOM with verification codes that are compatible with the Java tool.
    Based on: https://github.com/spdx/tools/blob/bc35e25/TestFiles/spdx-parser-source/org/spdx/rdfparser/VerificationCodeGenerator.java

    Args:
        software (List[Software]): The software entries to include in the generated verification code.

    Returns:
        Tuple[str, List[str]]: A tuple consisting of the generated package verification code, and a list of
        skipped file names.
    """
    skippedFileNames: List[str] = []
    fileNameAndChecksums: List[str] = collect_file_data(software)
    fileNameAndChecksums.sort()
    m = hashlib.sha1()
    for entry in fileNameAndChecksums:
        m.update(bytes(entry, "utf-8"))
    return m.hexdigest, skippedFileNames


def collect_file_data(software: List[Software]) -> List[str]:
    """Collect file checksums and paths as done by the Java SPDX Tools.

    Args:
        software (List[Software]): The software entries to include in the collected data.

    Returns:
        List[str]: A list of checksums and file paths, in the form "<checksum>||<filepath>".
    """
    file_data: List[str] = []
    for sw in software:
        # lower case sha1 hash without a leading 0x prefix
        checksum = sw.sha1.lower()
        for fpath in sw.installPath:
            # GitHub spdx/tools uses this format for checksumming file data: checksumValue+"||"+filePath+END_OF_LINE_CHAR
            file_data.append(checksum + "||" + normalize_file_path(fpath) + "\n")
    return file_data


def normalize_file_path(nonNormalizedFilePath: str) -> str:
    """Normalization of file paths as done by Java SPDX Tools (not part of the SPDX specification).

    NOTE: The Java GitHub spdx/tools seems to have several bugs. It uses a './' replacement method
    that is too liberal and will mess up directory names ending in a '.'; it also does not appear
    to correctly remove the previously added part of the a path when it encounters a '..' relative
    directory.

    Args:
        nonNormalizedFilePath (str): The file path to normalize.

    Returns:
        str: A normalized file path.
    """
    # replace backslashes with forward slash
    filePath = nonNormalizedFilePath.replace("\\", "/").strip()
    # this splits the filepath, and removes instances of "./"
    filePathParts = pathlib.Path(filePath).parts
    # move up to higher directories
    if ".." in filePathParts:
        normalizedFilePathParts = []
        for _, part in enumerate(filePathParts):
            if part == "..":
                # go up a directory by removing the last entry from filePathParts
                # going above the last directory could be considered an error
                if len(normalizedFilePathParts) > 0:
                    normalizedFilePathParts.pop()
            else:
                normalizedFilePathParts.append(part)
        filePath = "/".join(normalizedFilePathParts)
    filePath = "./" + filePath
    return filePath



from typing import Optional

import numbat

import surfactant.plugin
from surfactant.sbomtypes import SBOM


@surfactant.plugin.hookimpl
def write_sbom(sbom: SBOM, outfile) -> None:
    db = numbat.SourcetrailDB.open(outfile.name, clear=True)
    soft_ids = {}
    for soft in sbom.software:
        if soft.name:
            soft_id = db.record_class(name=soft.name)
        elif len(soft.fileName) > 0:
            soft_id = db.record_class(name=soft.fileName[0])
        else:
            continue
        soft_ids[soft.UUID] = soft_id
        for path in soft.installPath:
            db.record_field(name=path, parent_id=soft_id)
        for name in soft.fileName:
            db.record_method(name=name, parent_id=soft_id)
        for meta in soft.metadata:
            if "installPathSymlinks" in meta:
                for path in meta["installPathSymlinks"]:
                    typedef_id = db.record_typedef_node(name=path)
                    db.record_ref_usage(typedef_id, soft_id)
            if "fileNameSymlinks" in meta:
                for path in meta["fileNameSymlinks"]:
                    typedef_id = db.record_typedef_node(name=path)
                    db.record_ref_usage(typedef_id, soft_id)
    for relation in sbom.relationships:
        if relation.relationship == "Uses":
            db.record_ref_usage(soft_ids[relation.xUUID], soft_ids[relation.yUUID])
    db.commit()
    db.close()


@surfactant.plugin.hookimpl
def short_name() -> Optional[str]:
    return "sourcetrail_db"



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import sys
from typing import Any, List, Optional

import pluggy
from loguru import logger

from surfactant.configmanager import ConfigManager
from surfactant.plugin import hookspecs


def _register_plugins(pm: pluggy.PluginManager) -> None:
    # pylint: disable=import-outside-toplevel
    # don't want all these imports as part of the file-level scope
    from surfactant.filetypeid import id_extension, id_hex, id_magic
    from surfactant.infoextractors import (
        a_out_file,
        coff_file,
        docker_image,
        elf_file,
        java_file,
        js_file,
        mach_o_file,
        native_lib_file,
        ole_file,
        pe_file,
    )
    from surfactant.input_readers import cytrics_reader
    from surfactant.output import (
        csv_writer,
        cyclonedx_writer,
        cytrics_writer,
        spdx_writer,
    )
    from surfactant.relationships import (
        dotnet_relationship,
        elf_relationship,
        java_relationship,
        pe_relationship,
    )

    internal_plugins = (
        id_magic,
        id_hex,
        id_extension,
        a_out_file,
        coff_file,
        docker_image,
        elf_file,
        java_file,
        mach_o_file,
        js_file,
        pe_file,
        ole_file,
        dotnet_relationship,
        elf_relationship,
        java_relationship,
        pe_relationship,
        csv_writer,
        cytrics_writer,
        cyclonedx_writer,
        spdx_writer,
        cytrics_reader,
        native_lib_file,
    )
    for plugin in internal_plugins:
        pm.register(plugin)


def set_blocked_plugins(pm: pluggy.PluginManager) -> None:
    """Gets the current list of blocked plugins from the config manager, then blocks and unregisters them with the plugin manager."""
    config_manager = ConfigManager()

    # Retrieve the current list of blocked plugins
    current_blocked_plugins = config_manager.get("core", "disable_plugins", [])
    for plugin_name in current_blocked_plugins:
        # Check if the plugin is already blocked
        if pm.is_blocked(plugin_name):
            logger.info(f"Plugin '{plugin_name}' is already disabled.")
            continue

        # Unregister the plugin
        plugin = pm.unregister(name=plugin_name)
        if plugin is None:
            logger.info(f"Disabled plugin '{plugin_name}' not found.")
            continue

        # Block the plugin to prevent future registration
        pm.set_blocked(plugin_name)


def get_plugin_manager() -> pluggy.PluginManager:
    pm = pluggy.PluginManager("surfactant")
    pm.add_hookspecs(hookspecs)
    pm.load_setuptools_entrypoints("surfactant")
    _register_plugins(pm)
    set_blocked_plugins(pm)
    pm.check_pending()
    return pm


def is_hook_implemented(pm: pluggy.PluginManager, plugin: object, hook_name: str) -> bool:
    """
    Checks if a specific hook is implemented by a given plugin.

    Args:
        pm (pluggy.PluginManager): The plugin manager instance.
        plugin (object): The plugin object to check.
        hook_name (str): The name of the hook to check for implementation.

    Returns:
        bool: True if the hook is implemented by the plugin, False otherwise.
    """
    hook_callers = pm.get_hookcallers(plugin)
    if hook_callers:
        for hook_caller in hook_callers:
            if hook_caller.name == hook_name:
                return True
    return False


def print_plugins(pm: pluggy.PluginManager) -> None:
    print("PLUGINS")
    for plugin in pm.get_plugins():
        plugin_name = pm.get_name(plugin) if pm.get_name(plugin) else ""
        print(f"\t> name: {plugin_name}")
        print(f"\t  canonical name: {pm.get_canonical_name(plugin)}")

        short_name = None

        # Check if the plugin has implemented the short_name hook
        hook_impl = is_hook_implemented(pm, plugin, "short_name")

        if hook_impl:
            short_name = plugin.short_name()  # Get the short name

        print(f"\t  short name: {short_name}\n")


def find_io_plugin(pm: pluggy.PluginManager, io_format: str, function_name: str) -> Optional[Any]:
    """
    Finds and returns a plugin that matches the specified input/output format and has the desired function.

    Args:
        pm (pluggy.PluginManager): The plugin manager instance.
        io_format (str): The name that the plugin is registered as.
        function_name (str): The name of the function.

    Returns:
        Optional[Any]: The found plugin instance that matches the specified `io_format` and
                        implements the `function_name`, or `None` if no such plugin is found.
                        If no plugin is found, an error is logged, and the program exits.

    Raises:
        SystemExit: If no plugin matching the criteria is found, the function logs an error message
                     and exits the program.
    """

    found_plugin = pm.get_plugin(io_format)

    if found_plugin is None:
        for plugin in pm.get_plugins():
            try:
                if plugin.short_name().lower() == io_format.lower() and hasattr(
                    plugin, function_name
                ):
                    found_plugin = plugin
                    break
            except AttributeError:
                pass

    if found_plugin is None:
        logger.error(f'No "{function_name}" plugin for format "{io_format}" found')
        sys.exit(1)

    return found_plugin


def find_plugin_by_name(pm: pluggy.PluginManager, name: str) -> Optional[Any]:
    """
    Finds a plugin by matching the given name against the plugin's registered name,
    canonical name, and its short name (if applicable).

    Args:
        pm (pluggy.PluginManager): The plugin manager instance.
        name (str): The name to match against the plugin's registered, canonical, and short names.

    Returns:
        Optional[Any]: The matched plugin instance or None if no match is found.
    """
    # Convert the name to lowercase for case-insensitive comparison
    name_lower = name.lower()

    for plugin in pm.get_plugins():
        # Get the registered and canonical names
        registered_name = pm.get_name(plugin).lower()
        canonical_name = pm.get_canonical_name(plugin).lower()

        # Check if the plugin has implemented the short_name hook
        short_name = None
        hook_impl = is_hook_implemented(pm, plugin, "short_name")

        if hook_impl:
            short_name = plugin.short_name()  # Get the short name

        # Convert short_name to lowercase if it exists
        short_name_lower = short_name.lower() if short_name else None

        # Check if any of the names match the given name
        if name_lower in (registered_name, canonical_name, short_name_lower):
            return plugin

    return None


def call_init_hooks(
    pm: pluggy.PluginManager, hook_filter: List[str] = None, command_name: str = None
) -> None:
    """
    Call the initialization hook for plugins that implement it.

    Args:
        pm (pluggy.PluginManager): The plugin manager instance.
        hook_filter (List[str]): A list of hook names to filter which plugins get initialized.
        command_name (str): The name of the command invoking the initialization.
    """
    for plugin in pm.get_plugins():
        if is_hook_implemented(pm, plugin, "init_hook"):
            # Check if the plugin implements any of the hooks in the filter
            if hook_filter:
                if not any(is_hook_implemented(pm, plugin, hook) for hook in hook_filter):
                    continue
            plugin.init_hook(command_name=command_name)



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import json
import pathlib
import tarfile
from enum import Enum, auto
from typing import Optional

import surfactant.plugin
from surfactant.infoextractors.coff_file import COFF_MAGIC_TARGET_NAME


class ExeType(Enum):
    ELF = auto()
    PE = auto()
    OLE = auto()
    JAVA_MACHOFAT = auto()
    MACHO32 = auto()
    MACHO64 = auto()


def is_docker_archive(filepath: str) -> bool:
    # pylint: disable=too-many-return-statements
    with tarfile.open(filepath) as tar:
        try:
            manifest_info = tar.getmember("manifest.json")
            if not manifest_info.isfile():
                return False
            with tar.extractfile(manifest_info) as manifest_file:
                manifest = json.load(manifest_file)
                # There's one entry in the list for each image
                if not isinstance(manifest, list):
                    return False
                for data in manifest:
                    # Just check if this data member exists
                    _ = tar.getmember(data["Config"])
                    # Now check that each of the layers exist
                    for layer in data["Layers"]:
                        _ = tar.getmember(layer)
                # Everything seems to exist and be in order; this is most likely a Docker archive
                return True
        except KeyError:
            return False


@surfactant.plugin.hookimpl(tryfirst=True)
def identify_file_type(filepath: str) -> Optional[str]:
    # pylint: disable=too-many-return-statements
    try:
        with open(filepath, "rb") as f:
            magic_bytes = f.read(265)
            if magic_bytes[:4] == b"\x7fELF":
                return "ELF"
            if magic_bytes[:2] == b"MZ":
                # Several file types start with the same `MZ` signature, so we need to handle them.
                # Regardless, the initial header (may) contain a pointer to an additional COFF
                # header, so we look for that as the first step.
                coff_addr = (
                    int.from_bytes(magic_bytes[0x3C:0x40], byteorder="little", signed=False)
                    & 0xFFFF
                )

                # Check to see if the coff_addr is still within the initial read; if not, we read up
                # to where it is.
                if coff_addr > len(magic_bytes):
                    magic_bytes += f.read(coff_addr + 4 - len(magic_bytes))

                # If coff_addr is still longer than what has been read so far, it points off the end
                # of the file, so the file is either malformed or something else is up.
                if coff_addr + 4 > len(magic_bytes):
                    return "Malformed PE"

                if magic_bytes[coff_addr : coff_addr + 4] != b"PE\x00\x00":
                    return "DOS"

                # Check for the linux kernel header at 0x202 (may require a second read)
                if len(magic_bytes) < 0x206:
                    magic_bytes += f.read(265)

                if magic_bytes[0x202:0x206] == b"HdrS":
                    return "Linux Kernel Image"

                # Otherwise, call it a PE and be done with it.
                return "PE"

            if magic_bytes[:8] == b"\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1":
                # MSI (install), MSP (patch), MST (transform), and MSM (merge) files are all types of OLE files
                # the root storage object CLSID is used to identify what it is (+ file extension)
                return "OLE"
            # Microsoft CAB files
            if magic_bytes[:4] == b"MSCF":
                return "MSCAB"
            # InstallShield CAB files
            if magic_bytes[:4] == b"ISc(":
                return "ISCAB"
            # For gzipped data, also filter by extension to avoid huge number of entries with limited info
            if magic_bytes[:2] == b"\x1f\x8b" and "".join(
                pathlib.Path(filepath).suffixes
            ).lower() in [
                ".tar.gz",
                ".cab.gz",
            ]:
                if is_docker_archive(filepath):
                    return "DOCKER_GZIP"
                return "GZIP"
            if magic_bytes[257:265] == b"ustar\x0000" or magic_bytes[257:265] == b"ustar  \x00":
                if is_docker_archive(filepath):
                    return "DOCKER_TAR"
                return "TAR"
            if magic_bytes[:4] in [b"PK\x03\x04", b"PK\x05\x06", b"PK\x07\x08"]:
                suffix = pathlib.Path(filepath).suffix.lower()
                if suffix in [".zip", ".zipx"]:
                    return "ZIP"
                # Java archive files of various types
                if suffix == ".jar":
                    return "JAR"
                if suffix == ".war":
                    return "WAR"
                if suffix == ".ear":
                    return "EAR"
                # Android packages
                if suffix == ".apk":
                    return "APK"
                # iOS/iPad applications
                if suffix == ".ipa":
                    return "IPA"
                # Windows app package
                if suffix == ".msix":
                    return "MSIX"
            # Magic for Java and Mach-O FAT Binary are the same
            if magic_bytes[:4] == b"\xca\xfe\xba\xbe":
                # Distinguish them the same way file magic and Apple do
                # https://opensource.apple.com/source/file/file-80.40.2/file/magic/Magdir/cafebabe.auto.html
                # https://github.com/file/file/blob/master/magic/Magdir/cafebabe
                if int.from_bytes(magic_bytes[4:8], byteorder="big", signed=False) <= 30:
                    return "MACHOFAT"
                return "JAVACLASS"
            if magic_bytes[:4] == b"\xbe\xba\xfe\xca":
                return "MACHOFAT"
            if magic_bytes[:4] in [b"\xca\xfe\xba\xbf", b"\xbf\xba\xfe\xca"]:
                return "MACHOFAT64"
            # Apple fat EFI binaries
            if magic_bytes[:4] == b"\x0e\xf1\xfa\b9":
                return "EFIFAT"
            # NOTE: the MACH032 and MACHO64 (normal byte order) magic may be located
            # at offset 0x1000 in some files
            if magic_bytes[:4] in [b"\xfe\xed\xfa\xce", b"\xce\xfa\xed\xfe"]:
                return "MACHO32"
            if magic_bytes[:4] in [b"\xfe\xed\xfa\xcf", b"\xcf\xfa\xed\xfe"]:
                return "MACHO64"
            # https://releases.llvm.org/2.7/docs/BitCodeFormat.html#magic
            if magic_bytes[:4] == b"\xde\xc0\x17\x0b":
                return "LLVM_BITCODE"
            if magic_bytes[:4] == b"BC\xc0\xde":
                return "LLVM_IR"
            # Need to check both small and big endian for a.out
            a_out_magic = [0x111, 0x108, 0x107, 0x0CC, 0x10B]
            if (
                int.from_bytes(magic_bytes[:4], byteorder="big", signed=False) & 0xFFFF
                in a_out_magic
            ):
                return "A.OUT big"
            if (
                int.from_bytes(magic_bytes[:4], byteorder="little", signed=False) & 0xFFFF
                in a_out_magic
            ):
                return "A.OUT little"
            if (
                int.from_bytes(magic_bytes[:2], byteorder="little", signed=False)
                in COFF_MAGIC_TARGET_NAME
            ):
                return "COFF"
            # XCOFF:
            # https://www.ibm.com/docs/en/aix/7.3?topic=formats-xcoff-object-file-format
            if magic_bytes[:2] == "\x1d\x00":
                return "XCOFF32"
            if magic_bytes[:2] == "\xf7\x01":
                return "XCOFF64"
            # ECOFF:
            # https://web.archive.org/web/20160305114748/http://h41361.www4.hp.com/docs/base_doc/DOCUMENTATION/V50A_ACRO_SUP/OBJSPEC.PDF
            if magic_bytes[:2] in ("\x83\x01", "\x88\x01", "\x8f\x01"):
                return "ECOFF"
            # AR:
            # https://www.garykessler.net/library/file_sigs.html
            if magic_bytes[:8] == b"!<arch>\n":
                return "AR_LIB"
            # OMF:
            # https://github.com/file/file/blob/c8bba134ac1f3c9f5/magic/Magdir/msvc#L22
            if (
                int.from_bytes(magic_bytes[0:4], byteorder="big", signed=False) & 0xFF0F80FF
            ) == 0xF00D0000:
                return "OMF_LIB"
            # zlib:
            # https://www.rfc-editor.org/rfc/rfc1950
            if len(magic_bytes) >= 2:
                cmf = magic_bytes[0]
                flg = magic_bytes[1]
                cm = cmf & 0x0F
                if cm == 8:
                    if (cmf * 256 + flg) % 31 == 0:
                        return "ZLIB"
            return None
    except FileNotFoundError:
        return None



import json
import uuid as uuid_module
from collections import deque
from typing import Dict, List, Tuple

import click
from loguru import logger

from surfactant.configmanager import ConfigManager
from surfactant.plugin.manager import find_io_plugin, get_plugin_manager
from surfactant.sbomtypes._relationship import Relationship
from surfactant.sbomtypes._sbom import SBOM
from surfactant.sbomtypes._system import System


@click.argument("sbom_outfile", envvar="SBOM_OUTPUT", type=click.File("w"), required=True)
@click.argument("input_sboms", type=click.File("r"), required=True, nargs=-1)
@click.option(
    "--config_file",
    type=click.File("r"),
    required=False,
    help="Config file for controlling some aspects of the merged SBOM, primarily the creation of a new top-level system object (settings here will typically take precedence over command line options)",
)
@click.option(
    "--output_format",
    is_flag=False,
    default=ConfigManager().get(
        "core", "output_format", fallback="surfactant.output.cytrics_writer"
    ),
    help="SBOM output format, options=surfactant.output.[cytrics|csv|spdx]_writer",
)
@click.option(
    "--input_format",
    is_flag=False,
    default="surfactant.input_readers.cytrics_reader",
    help="SBOM input format, assumes that all input SBOMs being merged have the same format, options=surfactant.input_readers.[cytrics|cyclonedx|spdx]_reader",
)
@click.option(
    "--system_uuid",
    is_flag=False,
    help="System UUID to use for relationships to a top-level system object",
)
@click.option(
    "--system_relationship",
    is_flag=False,
    default="Contains",
    show_default=True,
    help="Relationship type between merged SBOM contents to a top-level system object",
)
@click.option(
    "--add_system/--no_add_system",
    default=False,
    show_default=True,
    help="Create a top-level system entry for tying together the merged SBOM components. When disabled, relationships will still be created to a provided system UUID",
)
@click.command("merge")
# pylint: disable-next=too-many-positional-arguments
def merge_command(
    input_sboms,
    sbom_outfile,
    config_file,
    output_format,
    input_format,
    system_uuid,
    system_relationship,
    add_system,
):
    """Merge two or more INPUT_SBOMS together into SBOM_OUTFILE.

    An optional CONFIG_FILE can be supplied to specify a root system entry
    """
    pm = get_plugin_manager()
    output_writer = find_io_plugin(pm, output_format, "write_sbom")
    input_reader = find_io_plugin(pm, input_format, "read_sbom")

    sboms = []
    for sbom in input_sboms:
        sboms.append(input_reader.read_sbom(sbom))
    config = None
    if config_file:
        config = json.load(config_file)
    merge(sboms, sbom_outfile, config, output_writer, system_uuid, system_relationship, add_system)


# pylint: disable-next=too-many-positional-arguments
def merge(
    input_sboms,
    sbom_outfile,
    config,
    output_writer,
    system_uuid=None,
    system_relationship="Contains",  # remember: keep in-sync with click arg default
    add_system=False,
):
    """Merge two or more SBOMs."""
    merged_sbom = input_sboms[0]
    for sbom_m in input_sboms[1:]:
        merged_sbom.merge(sbom_m)

    rel_graph = construct_relationship_graph(merged_sbom)
    roots = get_roots_check_cycles(rel_graph)

    # Check if provided UUID for a system object already exists to avoid creating a duplicate
    if config and "system" in config:
        if "UUID" in config["system"]:
            for s in merged_sbom.systems:
                if config["system"]["UUID"] == s.UUID:
                    add_system = False
                    break
    # Even if not adding the system, create a dummy/placeholder with the UUID for creating relationships
    system, using_random_uuid = create_system_object(merged_sbom, config, system_uuid)
    if add_system:
        merged_sbom.systems.append(system)

    # Add a system relationship to each root software/systems entry identified
    if not using_random_uuid or add_system:
        if config and "systemRelationship" in config:
            system_relationship = config["systemRelationship"]
        for r in roots:
            merged_sbom.relationships.add(
                Relationship(xUUID=system.UUID, yUUID=r, relationship=system_relationship)
            )
    else:
        logger.warning(
            "No top-level system relationships added; enable the add system option to randomly generate a UUID, or specify a system UUID"
        )

    output_writer.write_sbom(merged_sbom, sbom_outfile)


def construct_relationship_graph(sbom: SBOM):
    """Function to get create a relationship graph of systems and software within an sbom

    Args:
        sbom (SBOM): The sbom to generate relationship graph from.
    """
    # construct a graph for adding a system relationship to all root software entries
    rel_graph: Dict[str, List[str]] = {}
    # add all UUIDs as nodes in the graph
    for system in sbom.systems:
        rel_graph[system.UUID] = []
    for sw in sbom.software:
        rel_graph[sw.UUID] = []
    # iterate through all relationships, adding edges to the adjacency list
    for rel in sbom.relationships:
        if rel.xUUID not in rel_graph or rel.yUUID not in rel_graph:
            logger.error(
                f"Either the xUUID or yUUID for the relationship does not exist in the graph: {rel = }"
            )
            continue
        # consider also including relationship type for the edge
        # treat as directed graph, with inverted edges (pointing to parents) so dfs will eventually lead to the root parent node for a (sub)graph
        rel_graph[rel.yUUID].append(rel.xUUID)
    return rel_graph


def get_roots_check_cycles(rel_graph):
    """Function to get roots of the sbom and check for circular dependencies

    Args:
        rel_graph: The relationship graph for an sbom.
    """
    visited = set()
    roots = set()
    rootFound = set()
    recursionStack = deque()

    def dfs(rel):
        recursionStack.append(rel)
        # if the node is already visited, no revisiting required
        if rel in visited:
            recursionStack.pop()
            return rel in rootFound
        # mark the node as visited
        visited.add(rel)
        # the node has no parents, it is a root
        if not rel_graph[rel]:
            roots.add(rel)
            rootFound.add(rel)
            recursionStack.pop()
            return True
        cycle = False
        # node is not a root, move on to parents
        for parent in rel_graph[rel]:
            # detect cycles
            if parent in recursionStack:
                logger.info(f"CYCLE DETECTED: {parent} {rel}")
                cycle = True
            if dfs(parent):
                rootFound.add(rel)
        # if there was a cycle, and none of the parents led to a definite root node
        if cycle and rel not in rootFound:
            logger.info(f"CYCLE AND NO ROOT FOUND, SETTING {rel} AS THE ROOT")
            roots.add(rel)
            rootFound.add(rel)
        recursionStack.pop()
        return rel in rootFound

    for rel in rel_graph:
        dfs(rel)
    logger.info(f"ROOTS: {roots}")
    return roots


def create_system_object(sbom: SBOM, config=None, system_uuid=None) -> Tuple[System, bool]:
    """Function to create an accurate system object

    Positional arguments:
        sbom (SBOM): The SBOM the system object is being created for.
        config: The user specified config json (Optional).

    Returns:
        Tuple[System, bool]: The created system object and a boolean indicating if a random UUID was used.
    """

    system = {}
    using_random_uuid = False
    if config and "system" in config:
        system = config["system"]

    # system_uuid supplied via command line overrides config file UUID
    if system_uuid:
        system["UUID"] = system_uuid
    elif "UUID" not in system:
        # No UUID, generate a random one...
        using_random_uuid = True
        system["UUID"] = str(uuid_module.uuid4())
    # check if the UUID appears valid based on the CyTRICS schema
    elif not sbom.is_valid_uuid4(system["UUID"]):
        invalid_uuid = system["UUID"]
        logger.error(f"Invalid uuid4 given ({invalid_uuid}) for the system")

    if "name" not in system:
        system["name"] = ""
    captureStart = -1
    captureEnd = -1
    for sw in sbom.software:
        if captureStart and sw.captureTime:
            if captureStart == -1 or sw.captureTime < captureStart:
                captureStart = sw.captureTime
        if captureEnd and sw.captureTime:
            if captureEnd == -1 or sw.captureTime > captureEnd:
                captureEnd = sw.captureTime
    if "captureStart" not in system or not system["captureStart"]:
        system["captureStart"] = captureStart
    if "captureEnd" not in system or not system["captureEnd"]:
        system["captureEnd"] = captureEnd
    return System(**system), using_random_uuid



# Copyright 2024 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT


import logging
from pathlib import Path

try:
    import ssdeep

    SSDEEP_PRESENT = True
except ImportError:
    SSDEEP_PRESENT = False
    logging.warning("SSDEEP is not installed, therefore those hashes will not be generated.")
import tlsh

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


def do_tlsh(bin_data: bytes):
    return tlsh.hash(bin_data)


def do_ssdeep(bin_data: bytes):
    return ssdeep.hash(bin_data)


@surfactant.plugin.hookimpl(specname="extract_file_info")
def fuzzyhashes(sbom: SBOM, software: Software, filename: str, filetype: str):
    """
    Generate TLSH and potentially SSDEEP fuzzy hashes for the provided files.
    :param sbom(SBOM): The SBOM that the software entry/file is being added to. Can be used to add observations or analysis data.
    :param software(Software): The software entry associated with the file to extract information from.
    :param filename (str): The full path to the file to extract information from.
    :param filetype (str): File type information based on magic bytes.
    """

    hashdata = [(do_tlsh, "tlsh")]
    if SSDEEP_PRESENT:
        hashdata.append((do_ssdeep, "ssdeep"))
    # Validate the file path
    existing_data = {}
    filename = Path(filename)
    if not filename.exists():
        raise FileNotFoundError(f"No such file: '{filename}'")

    if all(hashname in existing_data for _, hashname in hashdata):
        # if everything is already in there, we just want to terminate without writing
        return None
    with open(filename, "rb") as f_bin:
        bin_data = f_bin.read()

    for hashfunc, hashname in hashdata:
        if hashname in existing_data:
            logging.info("Already %s hashed %s", hashname, filename.name)
        else:
            logging.info(
                "Found existing JSON file for %s but without '%s' key. Proceeding with hashing.",
                filename.name,
                hashname,
            )
            existing_data[hashname] = hashfunc(bin_data)
    return existing_data



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import json
from pathlib import Path

import binary2strings as b2s
from loguru import logger

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


@surfactant.plugin.hookimpl(specname="extract_file_info")
def extract_strings(sbom: SBOM, software: Software, filename: str, filetype: str):
    """
    Extract ASCII strings from a binary file using binary2strings.
    :param sbom(SBOM): The SBOM that the software entry/file is being added to. Can be used to add observations or analysis data.
    :param software(Software): The software entry associated with the file to extract information from.
    :param filename (str): The full path to the file to extract information from.
    :param filetype (str): File type information based on magic bytes.
    """
    shaHash = str(software.sha256)
    min_len = 4
    # Only parsing executable files
    if filetype not in ["ELF", "PE"]:
        pass

    filename = Path(filename)
    flist = []

    # Performing check to see if file has been analyzed already
    existing_json = None
    output_name = None
    for f in Path.cwd().glob("*_additional_metadata.json"):
        flist.append((f.stem).split("_")[0])
        if shaHash == (f.stem).split("_")[0]:
            existing_json = f
            output_name = f

    if existing_json:
        with open(existing_json, "r") as json_file:
            existing_data = json.load(json_file)
        if "strings" in existing_data:
            logger.info(f"Already extracted {filename.name}")
        else:
            logger.info(
                f"Found existing JSON file for {filename.name} but without 'strings' key. Proceeding with extraction."
            )
            # Add your extraction code here.
            if filename.name not in existing_data["filename"]:
                existing_data["filename"].append(filename.name)

            existing_data["strings"] = []
            # Extract and write strings using binary2strings
            with open(filename, "rb") as f_bin:
                data = f_bin.read()
                for string, _encoding, _span, _is_interesting in b2s.extract_all_strings(
                    data, only_interesting=True
                ):
                    if len(string) >= min_len:
                        existing_data["strings"].append(string)
            # Write the string_dict to the output JSON file
            with open(output_name, "w") as json_file:
                json.dump(existing_data, json_file, indent=4)

    else:
        # Validate the file path
        if not filename.exists():
            raise FileNotFoundError(f"No such file: '{filename}'")

        # Extract filename without extension
        output_path = Path.cwd() / f"{shaHash}_additional_metadata.json"
        string_dict = {}
        string_dict["sha256hash"] = shaHash
        string_dict["filename"] = [filename.name]
        string_dict["strings"] = []

        # Extract and write strings using binary2strings
        with open(filename, "rb") as f_bin:
            data = f_bin.read()
            for string, _encoding, _span, _is_interesting in b2s.extract_all_strings(
                data, only_interesting=True
            ):
                # you might adjust the condition below to filter strings based on your needs
                if len(string) >= min_len:
                    string_dict["strings"].append(string)

        # Write the string_dict to the output JSON file
        with open(output_path, "w") as json_file:
            json.dump(string_dict, json_file, indent=4)

        logger.info(f"Data written to {output_path}")



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from dataclasses import dataclass


@dataclass
class Relationship:
    xUUID: str
    yUUID: str
    relationship: str

    def __hash__(self) -> int:
        return hash(repr(self))


@dataclass
class StarRelationship:
    xUUID: str
    yUUID: str
    relationship: str

    def __hash__(self) -> int:
        return hash(repr(self))



# Copyright 2024 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import argparse
import json
import os
import re

# The following code adds the data present in additional_metadata.json files to an input
# sbom and outputs it at a new location.

#     - It uses the sha256hash field to perform linkages
#     - It does overwrite the output location

# Usage:
#     python3 scripts/merge_additional_metadata.py . sbom_without_metadata.json output_sbom_file.json


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "metadata_dir",
        help='The directory that contains the "additional metadata" files',
    )
    parser.add_argument("input_sbom", help="The SBOM for the additional metadata to be merged into")
    parser.add_argument("output_file", help="The output file")
    _args = parser.parse_args()
    return _args


if __name__ == "__main__":
    args = parse_args()
    with open(args.input_sbom) as f:
        sbom_data = json.load(f)
    lookup_table = {
        sbom_node["sha256"]: index for index, sbom_node in enumerate(sbom_data["software"])
    }
    for path in os.scandir(args.metadata_dir):
        if re.match("[a-z0-9]{64}_additional_metadata.json", path.name):
            with open(path) as f:
                additional_data = json.load(f)
            if additional_data["sha256hash"] in lookup_table:
                index = lookup_table[additional_data["sha256hash"]]
                if "metadata" not in sbom_data["software"][index]:
                    sbom_data["software"][index]["metadata"] = []
                sbom_data["software"][index]["metadata"].append(additional_data)
    with open(args.output_file, "w") as f:
        json.dump(sbom_data, f, indent=4)



import os
import platform
from pathlib import Path

import pytest

from surfactant.configmanager import ConfigManager


@pytest.fixture(name="config_manager")
def fixture_config_manager(tmp_path):
    # Use the tmp_path fixture for the temporary directory
    config_manager = ConfigManager(app_name="testapp", config_dir=tmp_path)
    yield config_manager
    # Cleanup after test
    ConfigManager.delete_instance("testapp")


def test_singleton(config_manager):
    config_manager2 = ConfigManager(app_name="testapp")
    assert config_manager is config_manager2


def test_set_and_get(config_manager):
    config_manager.set("Settings", "theme", "dark")
    theme = config_manager.get("Settings", "theme")
    assert theme == "dark"


def test_set_and_getitem(config_manager):
    config_manager.set("Settings", "theme", "dark")
    theme = config_manager["Settings"]["theme"]
    assert theme == "dark"


def test_createinstance_and_getitem(config_manager):
    config_manager.set("Settings", "theme", "dark")
    # ConfigManager instance accessed will be the same as the one created in the test fixture
    # so that the set value above will be present for testing
    settings_config = ConfigManager(app_name="testapp")["Settings"]
    assert settings_config
    assert "theme" in settings_config  # pylint: disable=unsupported-membership-test
    assert settings_config["theme"] == "dark"  # pylint: disable=unsubscriptable-object


def test_get_with_fallback(config_manager):
    fallback_value = "light"
    theme = config_manager.get("Settings", "theme", fallback=fallback_value)
    assert theme == fallback_value


def test_config_file_creation(config_manager):
    config_manager.set("Settings", "theme", "dark")
    assert config_manager.config_file_path.exists()


@pytest.mark.skipif(platform.system() != "Windows", reason="Test specific to Windows platform")
def test_windows_config_path():
    config_manager = ConfigManager(app_name="testapp")
    config_path = config_manager._get_config_file_path()  # pylint: disable=protected-access
    expected_config_dir = Path(os.getenv("APPDATA", str(Path("~\\AppData\\Roaming").expanduser())))
    assert expected_config_dir in config_path.parents
    assert config_path.parts[-2:] == ("testapp", "config.toml")
    # delete instance so other tests don't accidentally use it
    config_manager.delete_instance("testapp")


@pytest.mark.skipif(platform.system() == "Windows", reason="Test specific to Unix-like platforms")
def test_unix_config_path():
    config_manager = ConfigManager(app_name="testapp")
    config_path = config_manager._get_config_file_path()  # pylint: disable=protected-access
    expected_config_dir = Path(os.getenv("XDG_CONFIG_HOME", str(Path("~/.config").expanduser())))
    assert expected_config_dir in config_path.parents
    assert config_path.parts[-2:] == ("testapp", "config.toml")
    # delete instance so other tests don't accidentally use it
    config_manager.delete_instance("testapp")


@pytest.mark.skipif(platform.system() != "Windows", reason="Test specific to Windows platform")
def test_windows_data_dir_path():
    config_manager = ConfigManager(app_name="testapp")
    data_dir = config_manager.get_data_dir_path()  # pylint: disable=protected-access
    expected_data_dir = Path(os.getenv("LOCALAPPDATA", str(Path("~\\AppData\\Local").expanduser())))
    assert expected_data_dir in data_dir.parents
    assert data_dir.name == "testapp"
    # delete instance so other tests don't accidentally use it
    config_manager.delete_instance("testapp")


@pytest.mark.skipif(platform.system() == "Windows", reason="Test specific to Unix-like platforms")
def test_unix_data_dir_path():
    config_manager = ConfigManager(app_name="testapp")
    data_dir = config_manager.get_data_dir_path()  # pylint: disable=protected-access
    expected_data_dir = Path(os.getenv("XDG_DATA_HOME", str(Path("~/.local/share").expanduser())))
    assert expected_data_dir in data_dir.parents
    assert data_dir.name == "testapp"
    # delete instance so other tests don't accidentally use it
    config_manager.delete_instance("testapp")


def test_preserve_comments(config_manager):
    # Create a config file with some value and add in a comment line
    config_manager.set("Settings", "theme", "dark")
    with open(config_manager.config_file_path, "a") as configfile:
        configfile.write("\n# This is a comment\n")
    # Force reload of cached config in the ConfigManager
    config_manager._load_config()  # pylint: disable=protected-access
    assert config_manager.get("Settings", "theme") == "dark"

    # Set a new value to make ConfigManager to save an updated config file
    config_manager.set("Settings", "language", "en")
    with open(config_manager.config_file_path, "r") as configfile:
        content = configfile.read()
    assert "# This is a comment" in content


def test_multiple_instances(tmp_path):
    # Make sure two separate config managers can more or less co-exist peacefully
    config_manager1 = ConfigManager(app_name="testapp1", config_dir=tmp_path)
    config_manager2 = ConfigManager(app_name="testapp2", config_dir=tmp_path)
    config_manager1.set("Settings", "theme", "dark")
    config_manager2.set("Settings", "theme", "light")
    assert config_manager1.get("Settings", "theme") == "dark"
    assert config_manager2.get("Settings", "theme") == "light"
    assert config_manager1.config_file_path.exists()
    assert config_manager2.config_file_path.exists()
    assert config_manager1.config_file_path != config_manager2.config_file_path
    ConfigManager.delete_instance("testapp1")
    ConfigManager.delete_instance("testapp2")


if __name__ == "__main__":
    pytest.main()



import dataclasses
import pickle
from dataclasses import Field

from loguru import logger

from surfactant.configmanager import ConfigManager
from surfactant.sbomtypes._sbom import SBOM


class Cli:
    """
    A base class that implements the surfactant cli basic functionality

    Attributes:
        sbom: An internal record of sbom entries the class adds to as it finds more matches.
        subset: An internal record of the subset of sbom entries from the last cli find call.
        sbom_filename: A string value of the filename where the loaded sbom is stored.
        subset_filename: A string value of the filename where the current subset result from the "cli find" command is stored.
        match_functions: A dictionary of functions that provide matching functionality for given SBOM fields (i.e. uuid, sha256, installpath, etc)
        camel_case_conversions: A dictionary of string conversions from all lowercase to camelcase. Used to convert python click options to match the SBOM attribute's case

    """

    sbom: SBOM = None
    subset: SBOM = None
    sbom_filename: str
    subset_filename: str
    match_functions: dict
    camel_case_conversions: dict

    def __init__(self):
        self.sbom_filename = "sbom_cli"
        self.subset_filename = "subset_cli"
        # Create data directory
        self.data_dir = ConfigManager().get_data_dir_path()
        self.data_dir.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def serialize(sbom: SBOM) -> str:
        """Serializes a given sbom.

        Args:
            bom (SBOM): An instance of an SBOM to serialize

        Returns:
            bytes: A binary representation of the serialized SBOM.
        """
        # NOTE: python pickle cannot pickle MappingProxyType, which is inherently included in the Field type.
        # Pickling is much faster than converting to json or msgpack (see MR for timings: https://github.com/LLNL/Surfactant/pull/261
        # To workaround the issue we replace the metadata attribute of the Field type (which is default mappingproxytype) with an empty dict
        # Upon deserialization, we recreate the class with dataclasses.replace() to ensure the unpickled class instance is intact. Without this,
        # the class function to_json() and to_dict() does not work as the Field type is no longer recongized.
        if isinstance(sbom, SBOM):
            for _, v in sbom.__dataclass_fields__.items():
                if isinstance(v, Field):
                    v.metadata = {}
            return pickle.dumps(sbom)
        logger.error(f"Could not serialize sbom - {type(sbom)} is not of type SBOM")
        return None

    @staticmethod
    def deserialize(data) -> SBOM:
        """Deserializes the given data and saves them in the SBOM class instance

        Args:
            data (bytes): The data to deserialize into an SBOM type

        Returns:
            SBOM: An SBOM instance.
        """
        try:
            sbom = pickle.loads(data)
            return dataclasses.replace(
                sbom
            )  # Create a copy to repopulate anything that got messed up in serialization
        except pickle.UnpicklingError as e:
            logger.error(f"Could not deserialize sbom from given data - {e}")
            return None



from pathlib import Path

from surfactant.cmd.cli_commands.cli_base import Cli
from surfactant.plugin.manager import find_io_plugin, get_plugin_manager


class Save(Cli):
    """
    A class that implements the surfactant cli save functionality

    Attributes:
        output_format (str): The format of the sbom to be outputted
    """

    def __init__(self, *args, output_format, **kwargs):
        """Executes the load class constructor

        Args:
            output_format (str): The format of the sbom to be outputted
        """
        self.output_format = output_format
        super().__init__(*args, **kwargs)

    def execute(self, output_file):
        """Executes the main functionality of the load class

        Args:
            output_file:  The file to save the sbom to
        """
        pm = get_plugin_manager()
        output_writer = find_io_plugin(pm, self.output_format, "write_sbom")

        with open(Path(self.data_dir, self.sbom_filename), "rb") as f:
            data = f.read()
        self.sbom = Cli.deserialize(data)
        output_writer.write_sbom(self.sbom, output_file)



"""execinstaller.py"""

import sys
from json import dumps
from logging import DEBUG, basicConfig, exception, info
from os import getcwd, listdir, makedirs, mkdir, remove
from os.path import abspath, exists
from shutil import copy, copy2
from sys import argv
from time import sleep

from virtualbox import Session, VirtualBox
from virtualbox.library_base import VBoxError

# Configure logging output
basicConfig(format="%(message)s", encoding="utf-16", level=DEBUG)

# Configurable strings
SHARED = "vb"
HOST_SF = getcwd() + f"/{SHARED}"
UNFILTERED = "results.txt"
FILTERED = "filtered.txt"
ARGS_FILE = "args.txt"
FILE_LIST = "files.txt"
FILE_SIGNAL = "done.txt"
NOT_EXTRACTED = "UnextractedFilesList.txt"
EXTRACTED = "ExtractedFilesList.txt"
LATEST_SNAPSHOT = "TestState"

# Passed in command line arguments
args = {"-machine": "WinDev2307Eval", "-path": None, "-debug": "off"}


def parse_args(vbox: VirtualBox) -> None:
    """Reads the command line arguments and creates a dictionary with settings
    that affect the behavior of the script. If any argument in 'args' is None,
    the script prints out correct usage and exits. The user must specify the
    target vm name and the path of the installer.

    Args:
        vbox (VirtualBox): Primary VB object

    Returns:
        None.
    """

    # String that indicates correct cmd line argument usage
    flags = "-machine '...' -path '...' (optional) -license '...'"
    usage = "Usage: python execinstaller.py " + flags

    args["-license"] = "Stepper"

    # Parse cmdargs into a dictionary
    for i in range(2, len(argv), 2):
        if argv[i - 1] in args:
            args[argv[i - 1]] = argv[i]
        else:
            sys.exit(f"{argv[i - 1]} is not a valid argument.\n{usage}")

    # Ensure args have been set
    for key, value in args.items():
        if value is None:
            sys.exit(f"{key} is a required argument.\n{usage}")

    # Ensure that the user has selected a valid vm name
    names = [m.name for m in vbox.machines]

    if args["-machine"] not in names:
        sys.exit(f"'{args['-machine']}' isn't in your vm list: \n{names}")

    # Expand installer paths
    args["-path"] = abspath(args["-path"])
    args["-type"] = args["-path"].rpartition(".")[-1]


def prepare_vm(vbox: VirtualBox, session: Session) -> None:
    """Boots up the specified virtual machine and transfers the specified
    installer from the host machine.

    IMPORTANT: For this to work, the target machine must be in a SAVED state
    and NOT running on VirtualBox. This will allow the session to obtain the
    write lock. Also, waiting for the OS to fully boot is a complex task, so
    restoring the VM from a saved state greatly simplifies the process.

    Args:
        vbox (VirtualBox): Primary VB object
        session (Session): Session associated with vbox

    Returns:
        None.
    """

    # Create the shared folder on the host machine
    if not exists(HOST_SF):
        mkdir(HOST_SF)

    # Retrieve machine and boot it
    try:
        machine = vbox.find_machine(args["-machine"])
        progress = machine.launch_vm_process(session, "gui", [])
        progress.wait_for_completion()
    except VBoxError as err:
        exception(err)
        sys.exit("Ensure the VM is in a saved and non-running state.")

    # Create the shared folder on the guest machine's V: drive
    try:
        s_machine = session.machine
        s_machine.create_shared_folder(SHARED, HOST_SF, True, True, "Z:")
    except VBoxError:
        exception(f"The shared folder ./{SHARED}/ already exists")

    # Save settings on the machine
    session.machine.save_settings()


def deploy_installer() -> None:
    """Inserts intaller and required arguments into the shared folder. Once
    the VM has finished executing the installer, the 'results.txt' file is
    retrieved from the VM.

    Args:
        None.

    Returns:
        None.
    """

    try:
        # Insert arguments into shared folder
        with open(ARGS_FILE, "w", encoding="utf-8") as f_handle:
            f_handle.write(dumps(args, separators=(",", ":")))

        copy(ARGS_FILE, HOST_SF)
        remove(ARGS_FILE)
    except PermissionError as err:
        exception(err)
        sys.exit("Make sure the files are in the current directory!")

    try:
        # Deploy installer
        copy(args["-path"], HOST_SF)
    except PermissionError as err:
        exception(err)
        sys.exit("Make sure the files are in the current directory!")

    # Busy wait until the installer is done
    while not exists(f"{HOST_SF}/{UNFILTERED}"):
        sleep(0.1)

    # Retrieve results
    sleep(2)
    copy(f"{HOST_SF}/{UNFILTERED}", getcwd())
    remove(f"{HOST_SF}/{UNFILTERED}")


def cleanup(session: Session) -> None:
    """Cleans up the shared folder from extra files, if possible, then powers
    down the VM and restores it from the state it started as. The Snapshot
    should be called "TestState" for this to work, but it can be changed.

    Args:
        session (Session): Session associated with vbox

    Returns:
        None.
    """

    # Cleanup the shared folder
    for file in listdir(f"./{SHARED}"):
        try:
            remove(f"./{SHARED}/{file}")
        except IOError as err:
            exception(f"CANNOT DELETE ./{SHARED}/{file}: {str(err)}")

    # Let the user read VM output if debug is on
    if args["-debug"] == "on":
        input("ENTER when you are done reading VM output...")

    # Save settings and power down the machine
    session.machine.save_settings()
    progress = session.console.power_down()
    progress.wait_for_completion()

    # Get latest snapshot and restore it
    snapshot = session.machine.find_snapshot(LATEST_SNAPSHOT)
    progress = session.machine.restore_snapshot(snapshot)
    progress.wait_for_completion()

    session.unlock_machine()


def get_attributes() -> list:
    """Reads from results.txt and returns the list of files

    Args:
        None.

    Returns:
        list: contains the attributes for each file
    """

    # Ensure that results.txt has been moved
    if not exists(UNFILTERED):
        return []

    info(f"processing {UNFILTERED}...")

    # Load file data
    lines = None

    with open(UNFILTERED, "r", encoding="utf-16") as f_handle:
        lines = f_handle.readlines()

    remove(UNFILTERED)

    # get parallel lists ready, prevent reading an incomplete file
    nfiles = len(lines) // 7
    attr = []
    for _ in range(6):
        attr.append([""] * nfiles)

    for i in range(nfiles):
        idx = i * 7

        # Interpret the extension buffer correctly
        ext = str(lines[idx + 2]).strip("[] \t\r\n")
        attr[1][i] = "".join([chr(int(x)) for x in ext.split(", ")])

        # Construct the POSIX file name and fix the duplicate extension
        currfname = str(lines[idx + 1]).split(".", 1)[0].strip('" \t\r\n')
        currfname = "BAD:" if len(currfname) == 0 else currfname
        attr[0][i] = currfname.replace("\\\\", "/") + "." + attr[1][i]

        # Get the rest of the attributes from the file
        exename = str(lines[idx + 3]).rstrip().strip('"')
        attr[2][i] = "BAD" if len(exename) == 0 else exename
        attr[3][i] = int(str(lines[idx + 4]).rstrip())
        attr[4][i] = int(str(lines[idx + 5]).rstrip())
        attr[5][i] = int(str(lines[idx + 6]).rstrip())

    return attr


def analyze_results() -> list:
    """Analyzes the unfiltered file and filters out the irrelevant files.
    The newly created file lists files that should be extracted from the VM.
    Attribute indices: 0 = filename, 1 = extension, 2 = executable,
    3 = pid, 4 = irp_operation, 5 = change_code

    Args:
        None.

    Returns:
        list: contains the filtered list of files that should be extracted.
    """

    # Create auxillary lists
    attr = get_attributes()
    nfiles = len(attr[0])
    aux_attr = [[], [], [], [], [], []]

    # List of unwanted exe substrings
    badexes = ["msteams", "TiWorker", "MicrosoftEdge", "MoUsoCoreWorker"]
    badexes.extend(["DeviceCensus", "msedge", "TrustedInstaller", "UsoClient"])
    badexes.extend(["MpSigStub", "OneDrive", "RuntimeBroker", "TaskHost"])
    badexes.extend(["Notepad", "Microsoft.SharePoint", "taskhostw", "svchost"])
    badexes.extend(["SecurityHealthHost", "MpCmdRun", "conhost", "dllhost"])
    badexes.extend(["MoNotificationUx", "taskkill", "CompatTelRunner"])
    badexes.extend(["python"])

    # List of unwanted file substrings
    badfiles = ["Prefetch", "taskkill"]

    # Remove duplicates and bad substrings
    for i in range(nfiles):
        # Remove duplicates and bad files
        cond = (attr[0][i] in aux_attr[0]) or attr[0][i].startswith("BAD:")

        # Filter out bad executables and file reads
        cond = cond or attr[2][i].startswith("BAD") or (attr[4][i] == 1)

        # Filter out file deletions
        if cond or (attr[5][i] == 6):
            continue

        # Filter out blacklisted executables and files
        skip = False
        for badexe in badexes:
            if badexe in attr[2][i]:
                skip = True

        for badfile in badfiles:
            if badfile in attr[0][i]:
                skip = True

        if skip:
            continue

        # Add unique file to the lists
        for j in range(6):
            aux_attr[j].append(attr[j][i])

    if args["-debug"] == "off":
        return aux_attr[0]

    # Reassign list references
    attr = aux_attr

    # Keep intermediate file for debugging
    nfiles = len(attr[0])

    with open(FILTERED, "w", encoding="utf-8") as f_handle:
        for i in range(nfiles):
            f_handle.write(f"File {i}:\n")
            f_handle.write(f"Name:     {attr[0][i]}\n")
            f_handle.write(f"Exe File: {attr[2][i]}\n")
            f_handle.write(f"PID/ops:  {attr[3][i]} {attr[4][i]} {attr[5][i]}")
            f_handle.write("\n\n")

    return attr[0]


def recreate_dirs(files: list) -> None:
    """Asks the VM to extract files created by the installer. Files that are
    successfully extracted are listed in "finalpaths.txt" and those that
    were not successful are listed in "otherpaths.txt". The files are then
    placed into an empty "C" volume within their full install paths. One can
    then run Surfactant on the skeleton "C" directory.

    Args:
        files (list): The list of files that the VM will attempt to retrieve.

    Returns:
        None.
    """

    newdirs = []

    # Send the list of desired files to the vm
    with open(FILE_LIST, "w", encoding="utf-8") as f_handle:
        for file in files:
            # Edit file path for the VM file to be the file path for a C folder
            dirs = "./C/" + "/".join(file.split("/")[1:-1]) + "/"
            newdirs.append(dirs)

            # Add to the list of files. Remove mysterious 21 space characters
            f_handle.write(file[:-21].replace(chr(0), ""))
            f_handle.write("\n")

    copy(FILE_LIST, f"./{SHARED}/{FILE_LIST}")

    if args["-debug"] == "off":
        remove(FILE_LIST)

    # Wait for the vm to send back all of the files
    while not exists(f"./{SHARED}/{FILE_SIGNAL}"):
        sleep(0.5)

    remove(f"./{SHARED}/{FILE_LIST}")

    # Move files from the shared folder to their new path
    num_files = len(files)

    for i in range(num_files):
        # The correct path for the new file is just a variant of its name
        file = files[i][3:].replace("/", "!SEP!")
        newpath = f"./C/{files[i][3:]}"
        newpath = newpath.replace(chr(0), "")
        sharedfile = f"./{SHARED}/{file}".replace(chr(0), "")

        # Write to the list of unsuccessful files if the file doesn't exist
        if not exists(sharedfile):
            info(f"Failure: {newpath}")
            info(" --> [Errno 2] No such file or directory")

            with open(NOT_EXTRACTED, "a", encoding="utf-8") as f_handle:
                f_handle.write(f"C:{newpath[3:]}\n")

            continue

        try:
            # If the file exists, make the directories for it and copy it in
            makedirs(newdirs[i], exist_ok=True)
            copy2(sharedfile, newpath)
            info(f"Success: {newpath}")

            # Write to the list of successful files
            with open(EXTRACTED, "a", encoding="utf-8") as f_handle:
                f_handle.write(f"C:{newpath[3:]}\n")

            remove(sharedfile)
        except OSError as err:
            # Since we know the file exists, deal with access errors
            exception(f"Copy/Removal Failure: {newpath}")
            exception(" --> " + str(err))

    # Signal to vm that all files have been transferred
    remove(f"./{SHARED}/{FILE_SIGNAL}")


def main() -> None:
    """Handles command line parsing and executes each segment.

    Args:
        None.

    Returns:
        None.
    """

    # Initiate VirtualBox variables
    vbox = VirtualBox()
    session = Session()

    # Read command line arguments and start the VM
    parse_args(vbox)
    prepare_vm(vbox, session)

    # Skip over to cleanup if a fatal error happened when deploying installer
    try:
        # Send installer to VM
        deploy_installer()

        # Extract files from VM
        files = analyze_results()
        recreate_dirs(files)
    except OSError as err:
        exception(err)

    # Restore VM to previous state
    cleanup(session)


if __name__ == "__main__":
    main()



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import os
import pathlib
import tempfile

import pytest

from surfactant.cmd.generate import resolve_link

base_dir = pathlib.Path(__file__).parent.absolute()


def symlink(src, dst, target_is_directory):
    try:
        os.symlink(src, dst, target_is_directory)
    except FileExistsError:
        pass


def create_symlinks(temp_dir):
    # Make sure this is always the working directory
    os.chdir(base_dir)
    os.makedirs(os.path.join(temp_dir, "test_dir", "subdir"), exist_ok=True)
    os.chdir(os.path.join(temp_dir, "test_dir"))
    symlink("..", "parent", True)
    symlink("parent", "link_to_parent", True)
    symlink("/none/", "does_not_exist", True)
    symlink("does_not_exist", "link_to_non_existant", False)
    symlink("..", "subdir/parent", True)
    # Revert back to the original working directory
    os.chdir(base_dir)


@pytest.mark.skipif(os.name != "posix", reason="requires posix os")
def test_symlinks():
    with tempfile.TemporaryDirectory() as temp_dir:
        create_symlinks(temp_dir)
        base_path = os.path.realpath(os.path.join(temp_dir, "test_dir"))
        assert resolve_link(os.path.join(base_path, "parent"), base_path, base_path) == base_path
        assert (
            resolve_link(os.path.join(base_path, "link_to_parent"), base_path, base_path)
            == base_path
        )
        assert resolve_link(os.path.join(base_path, "does_not_exist"), base_path, base_path) is None
        assert (
            resolve_link(
                os.path.join(base_path, "subdir", "parent"),
                os.path.join(base_path, "subdir"),
                base_path,
            )
            == base_path
        )
        assert (
            resolve_link(os.path.join(base_path, "link_to_non_existant"), base_path, base_path)
            is None
        )


if __name__ == "__main__":
    test_symlinks()
